import multiprocessing
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from testing.testcases import LiveTornadoTestCase
from .editor_helper import EditorHelper
from document.ws_views import WebSocket


class OfflineTests(LiveTornadoTestCase, EditorHelper):
    """
    Tests in which two browsers collaborate and the connection is interrupted.
    """
    user = None
    TEST_TEXT = "Lorem ipsum dolor sit amet."
    MULTILINE_TEST_TEXT = "Lorem ipsum\ndolor sit amet."
    fixtures = [
        'initial_documenttemplates.json',
        'initial_styles.json',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        driver_data = cls.get_drivers(2)
        cls.driver = driver_data["drivers"][0]
        cls.driver2 = driver_data["drivers"][1]
        cls.client = driver_data["clients"][0]
        cls.client2 = driver_data["clients"][1]
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        cls.driver2.quit()
        super().tearDownClass()

    def setUp(self):
        self.user = self.create_user()
        self.login_user(self.user, self.driver, self.client)
        self.login_user(self.user, self.driver2, self.client2)
        self.doc = self.create_new_document()

    def tearDown(self):
        self.leave_site(self.driver)
        self.leave_site(self.driver2)

    def test_simple(self):
        """
        Test one client going offline in collaborative mode while both clients
        continue to write and whether documents are synched when user returns
        online.
        """
        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2,
            len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

    def test_too_many_diffs(self):
        """
        Test one client going offline in collaborative mode while both clients
        continue to write with the cionnected clients adding too many items to
        the history so that the server no longer can provide it with all
        missing steps. The client therefore needs to recretae the missing steps
        by itself.
        """

        # The history length stored by the server is shortened from 1000 to 1.
        WebSocket.history_length = 1

        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2,
            len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

        WebSocket.history_length = 1000

    def test_too_many_diffs_with_tracking(self):
        """
        Test one client going offline in collaborative mode while both clients
        continue to write with the cionnected clients adding too many items to
        the history so that the server no longer can provide it with all
        missing steps. The client therefore needs to recreate the missing steps
        by itself. The limit of steps is set so that tracking kicks in.
        """

        # The history length stored by the server is shortened from 1000 to 1.
        WebSocket.history_length = 1

        self.load_document_editor(self.driver, self.doc)
        self.load_document_editor(self.driver2, self.doc)

        self.add_title(self.driver)
        self.driver.find_element_by_class_name(
            'article-body'
        ).click()

        p1 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver, self.TEST_TEXT)
        )
        p1.start()

        # Wait for the first processor to write some text
        self.wait_for_doc_size(self.driver2, 34)

        # driver 2 sets tracking limit
        self.driver2.execute_script(
            'window.theApp.page.mod.collab.doc.trackOfflineLimit = 0'
        )

        # driver 2 goes offline
        self.driver2.execute_script(
            'window.theApp.page.ws.goOffline()'
        )

        self.driver2.find_element_by_class_name(
            'article-body'
        ).click()

        # Total: 25
        self.driver2.execute_script(
            'window.testCaret.setSelection(25,25)'
        )

        p2 = multiprocessing.Process(
            target=self.type_text,
            args=(self.driver2, self.TEST_TEXT)
        )
        p2.start()
        p1.join()
        p2.join()

        # driver 2 goes online
        self.driver2.execute_script(
            'window.theApp.page.ws.goOnline()'
        )

        self.wait_for_doc_sync(self.driver, self.driver2)

        self.assertEqual(
            len(self.TEST_TEXT) * 2,
            len(self.get_contents(self.driver))
        )

        self.assertEqual(
            self.get_contents(self.driver2),
            self.get_contents(self.driver)
        )

        dialogtitle = WebDriverWait(self.driver2, self.wait_time).until(
            EC.element_to_be_clickable((By.CLASS_NAME, "ui-dialog-title"))
        )

        assert dialogtitle.text == 'System message'
        self.driver2.find_element_by_css_selector(
            '.ui-dialog button.fw-orange.fw-button'
        ).click()

        change_tracking_boxes = self.driver2.find_elements_by_css_selector(
            '.margin-box.track'
        )
        self.assertEqual(
            len(change_tracking_boxes),
            1
        )

        WebSocket.history_length = 1000



from django.conf.urls import url

from . import views

urlpatterns = [
    url('^flatpage/$', views.flatpage, name='flatpage'),
    url('^configuration/$', views.configuration, name='configuration'),
]



from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from testing.testcases import LiveTornadoTestCase
from testing.selenium_helper import SeleniumHelper


class PreloginTest(LiveTornadoTestCase, SeleniumHelper):
    fixtures = [
        'initial_terms.json',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.base_url = cls.live_server_url
        driver_data = cls.get_drivers(1)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.driver.implicitly_wait(driver_data["wait_time"])
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        super().tearDownClass()

    def test_flatpage(self):
        self.driver.get(self.base_url + "/")
        self.driver.find_element(
            By.CSS_SELECTOR,
            "a[href='/pages/terms/']"
        ).click()
        h3 = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.CSS_SELECTOR, 'h3'))
        )
        self.assertEqual(
            h3.text,
            'Your Account and Documents on the Website'
        )
        self.driver.find_element(
            By.CSS_SELECTOR,
            "a[href='/pages/privacy/']"
        ).click()
        h3 = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, 'h3:nth-child(4)')
            )
        )
        self.assertEqual(
            h3.text,
            'B. Collecting personal information'
        )

    def test_language_switch(self):
        driver = self.driver
        driver.get(self.base_url + "/")
        self.driver.find_element(
            By.ID,
            "lang-selection"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "#lang-selection option[value=es]"
        ).click()
        self.assertEqual(
            self.driver.find_element(
                By.CSS_SELECTOR,
                "html[lang=es] h1.fw-login-title"
            ).text,
            'INICIAR SESIÓN'
        )
        self.assertEqual(
            self.driver.find_element(
                By.ID,
                "lang-selection"
            ).get_attribute('value'),
            'es'
        )
        self.driver.find_element(
            By.ID,
            "lang-selection"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "#lang-selection option[value=en]"
        ).click()
        self.assertEqual(
            self.driver.find_element(
                By.CSS_SELECTOR,
                "html[lang=en] h1.fw-login-title"
            ).text,
            'LOG IN'
        )



import os
import time
from tempfile import mkdtemp

from testing.testcases import LiveTornadoTestCase
from testing.selenium_helper import SeleniumHelper
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait, Select
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from django.conf import settings


class ExportTest(LiveTornadoTestCase, SeleniumHelper):
    """Test whether Fidus Writer exports files in all the formats.
    Note that it does not validate the export files."""
    fixtures = [
        'initial_documenttemplates.json',
        'initial_styles.json',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.base_url = cls.live_server_url
        cls.download_dir = mkdtemp()
        driver_data = cls.get_drivers(1, cls.download_dir)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.driver.implicitly_wait(driver_data["wait_time"])
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        os.rmdir(cls.download_dir)
        super().tearDownClass()

    def setUp(self):
        self.verificationErrors = []
        self.accept_next_alert = True
        self.user1 = self.create_user(
            username='Yeti',
            email='yeti@snowman.com',
            passtext='otter'
        )

    def tearDown(self):
        self.leave_site(self.driver)

    def test_export(self):
        self.driver.get(self.base_url)
        self.driver.find_element(By.ID, "id_login").send_keys("Yeti")
        self.driver.find_element(By.ID, "id_password").send_keys("otter")
        self.driver.find_element(By.ID, "login-submit").click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    ".new_document button"
                )
            )
        ).click()
        # Set copyright
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((
                By.CSS_SELECTOR,
                '#header-navigation > div:nth-child(3) > span'
            ))
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            (
                "#header-navigation > div:nth-child(3) > "
                "div > ul > li:nth-child(6) > span"
            )
        ).click()
        self.driver.find_element_by_css_selector(
            '.holder'
        ).send_keys('The Big Company')
        self.driver.find_element_by_css_selector(
            '.year'
        ).send_keys('2003')
        self.driver.find_element_by_css_selector(
            '.license'
        ).click()
        self.driver.find_element_by_css_selector(
            '.license > option:nth-child(4)'
        ).click()
        self.driver.find_element_by_css_selector(
            '.license-start'
        ).send_keys('2001-04-06')
        self.driver.find_element_by_css_selector(
            "[aria-describedby=configure-copyright] button.fw-dark"
        ).click()

        self.driver.find_element(By.CSS_SELECTOR, ".article-title").click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-title").send_keys(
            "Title"
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            "No styling"
        )
        self.driver.find_element(
            By.CSS_SELECTOR,
            "button[title=Strong]"
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            "strong"
        )
        self.driver.find_element(
            By.CSS_SELECTOR,
            "button[title=Strong]"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "button[title=Emphasis]"
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            "emph"
        )
        self.driver.find_element(
            By.CSS_SELECTOR,
            "button[title=Emphasis]"
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            Keys.ENTER
        )
        self.driver.find_element(By.CSS_SELECTOR, ".fa-list-ol").click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            "ordered list"
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            Keys.ENTER
        )
        self.driver.find_element(By.CSS_SELECTOR, ".fa-list-ul").click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            "bullet list"
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            Keys.ENTER
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            Keys.ENTER
        )
        self.driver.find_element(
            By.CSS_SELECTOR,
            "button[title=Blockquote]"
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            "block quote"
        )
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            Keys.ENTER
        )
        self.driver.find_element(By.CSS_SELECTOR, ".fa-link").click()
        self.driver.find_element(By.CSS_SELECTOR, ".link-title").click()
        self.driver.find_element(By.CSS_SELECTOR, ".link-title").send_keys(
            "Sports"
        )
        self.driver.find_element(By.CSS_SELECTOR, ".link").click()
        self.driver.find_element(By.CSS_SELECTOR, ".link").send_keys(
            "https://www.sports.com"
        )
        self.driver.find_element(By.CSS_SELECTOR, ".fw-dark").click()
        self.driver.find_element(By.CSS_SELECTOR, ".fa-asterisk").click()
        self.driver.find_element(
            By.CSS_SELECTOR, ".footnote-container > p").click()
        self.driver.find_element(
            By.CSS_SELECTOR, ".footnote-container > p").send_keys('A footnote')
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").click()
        self.driver.find_element(By.CSS_SELECTOR, ".fa-book").click()
        # click on 'Register new source' button
        register_new_source = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.CLASS_NAME, 'register-new-bib-source')
            )
        )
        register_new_source.click()

        # select source
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.ID, 'select-bibtype'))
        )

        # click on article
        Select(
            self.driver.find_element_by_id("select-bibtype")
        ).select_by_visible_text("Article")

        # fill the values
        title_of_publication = self.driver.find_element_by_css_selector(
            '.journaltitle .ProseMirror'
        )
        title_of_publication.click()
        title_of_publication.send_keys("My publication title")

        title = self.driver.find_element_by_css_selector(
            '.title .ProseMirror')
        title.click()
        title.send_keys("My title")

        author_firstName = self.driver.find_element_by_css_selector(
            '.author .given .ProseMirror')
        author_firstName.click()
        author_firstName.send_keys("John")

        author_lastName = self.driver.find_element_by_css_selector(
            '.family .ProseMirror')
        author_lastName.click()
        author_lastName.send_keys("Doe")

        publication_date = self.driver.find_element_by_css_selector(
            '.date .date'
        )
        publication_date.click()
        publication_date.send_keys("2012")

        # click on Submit button
        self.driver.find_element_by_xpath(
            '//*[contains(@class, "ui-button") and normalize-space()="Submit"]'
        ).click()

        # Wait for source to be listed
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    '#selected-cite-source-table .selected-source'
                )
            )
        )
        # click on Insert button
        self.driver.find_element_by_css_selector('.insert-citation').click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            'button[title="Horizontal line"]'
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, ".article-body").send_keys(
            Keys.DOWN
        )
        # Insert figure
        button = self.driver.find_element_by_xpath('//*[@title="Figure"]')
        button.click()

        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, "caption"))
        ).send_keys('Figure')

        # click on 'Insert image' button
        self.driver.find_element_by_id('insert-figure-image').click()

        upload_button = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (
                    By.XPATH,
                    '//*[normalize-space()="Add new image"]'
                )
            )
        )

        upload_button.click()

        # image path
        image_path = os.path.join(
            settings.PROJECT_PATH,
            'document/tests/uploads/image.png'
        )

        # in order to select the image we send the image path in the
        # LOCAL MACHINE to the input tag
        upload_image_url = WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located(
                (By.XPATH, '//*[@id="editimage"]/div[1]/input[2]')
            )
        )
        upload_image_url.send_keys(image_path)

        # click on 'Upload' button
        self.driver.find_element_by_xpath(
            '//*[contains(@class, "ui-button") and normalize-space()="Upload"]'
        ).click()

        # click on 'Use image' button
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (By.CSS_SELECTOR, '.fw-data-table i.fa-check')
            )
        )

        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Use image"]'
        ).click()

        # Set copyright info
        self.driver.find_element_by_css_selector(
            "div.figure-preview > div > span > i"
        ).click()
        self.driver.find_element_by_css_selector(
            (
                "body > div.ui-content_menu.ui-corner-all.ui-widget."
                "ui-widget-content.ui-front > div > div > ul > li:nth-child(1)"
            )
        ).click()
        self.driver.find_element_by_css_selector(
            '.holder'
        ).send_keys('Johannes Wilm')
        self.driver.find_element_by_css_selector(
            '.year'
        ).send_keys('1998')
        self.driver.find_element_by_css_selector(
            '.license'
        ).click()
        self.driver.find_element_by_css_selector(
            '.license > option:nth-child(2)'
        ).click()
        self.driver.find_element_by_css_selector(
            '.license-start'
        ).send_keys('1998-04-23')
        self.driver.find_element_by_css_selector(
            "[aria-describedby=configure-copyright] button.fw-dark"
        ).click()

        # click on 'Insert' button
        self.driver.find_element_by_css_selector("button.fw-dark").click()
        ActionChains(self.driver).send_keys(
            Keys.RIGHT
        ).perform()
        # Add table
        self.driver.find_element(By.CSS_SELECTOR, ".fa-table").click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "table.insert-table-selection tr:nth-child(2) > td:nth-child(2)"
        ).click()
        self.driver.find_element(By.CSS_SELECTOR, ".fw-dark").click()
        self.driver.find_element(
            By.CSS_SELECTOR, "tr:nth-child(1) > td:nth-child(1) > p").click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "tr:nth-child(1) > td:nth-child(1) > p").send_keys('one')
        self.driver.find_element(
            By.CSS_SELECTOR, "tr:nth-child(1) > td:nth-child(2) > p").click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "tr:nth-child(1) > td:nth-child(2) > p").send_keys('two')
        self.driver.find_element(
            By.CSS_SELECTOR, "tr:nth-child(2) > td:nth-child(1) > p").click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "tr:nth-child(2) > td:nth-child(1) > p").send_keys('three')
        self.driver.find_element(
            By.CSS_SELECTOR, "tr:nth-child(2) > td:nth-child(2)").click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "tr:nth-child(2) > td:nth-child(2)").send_keys('four')
        # Document with many features has been created let's see if we can
        # export it from the editor.

        # HTML
        self.driver.find_element(
            By.CSS_SELECTOR,
            '.header-nav-item[title="Export of the document contents"]'
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="HTML"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(os.path.join(
            self.download_dir, 'title.html.zip'))
        os.remove(os.path.join(self.download_dir, 'title.html.zip'))

        # EPUB
        self.driver.find_element(
            By.CSS_SELECTOR,
            '.header-nav-item[title="Export of the document contents"]'
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Epub"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(os.path.join(self.download_dir, 'title.epub'))
        os.remove(os.path.join(self.download_dir, 'title.epub'))

        # LaTeX
        self.driver.find_element(
            By.CSS_SELECTOR,
            '.header-nav-item[title="Export of the document contents"]'
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="LaTeX"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(os.path.join(
            self.download_dir, 'title.latex.zip'))
        os.remove(os.path.join(self.download_dir, 'title.latex.zip'))

        # JATS
        self.driver.find_element(
            By.CSS_SELECTOR,
            '.header-nav-item[title="Export of the document contents"]'
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="JATS"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(os.path.join(
            self.download_dir, 'title.jats.zip'))
        os.remove(os.path.join(self.download_dir, 'title.jats.zip'))

        # DOCX
        self.driver.find_element(
            By.CSS_SELECTOR,
            '.header-nav-item[title="Export of the document contents"]'
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Classic (DOCX)"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(os.path.join(self.download_dir, 'title.docx'))
        os.remove(os.path.join(self.download_dir, 'title.docx'))

        # ODT
        self.driver.find_element(
            By.CSS_SELECTOR,
            '.header-nav-item[title="Export of the document contents"]'
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Free (ODT)"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(os.path.join(self.download_dir, 'title.odt'))
        os.remove(os.path.join(self.download_dir, 'title.odt'))

        # Save a revision
        self.driver.find_element(
            By.CSS_SELECTOR, '.header-nav-item[title="File handling"]').click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Save revision"]'
        ).click()
        self.driver.find_element_by_css_selector(
            '.revision-note'
        ).send_keys('First revision')
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Save"]'
        ).click()

        # Exit the editor
        self.driver.find_element(
            By.ID,
            "close-document-top"
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.ID, 'preferences-btn'))
        )

        # Export from overview
        self.driver.find_element_by_css_selector(
            'tr:nth-child(1) > td > label'
        ).click()

        # Native
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, '.dt-bulk-dropdown'))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Export selected as FIDUS"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(
            os.path.join(self.download_dir, 'title.fidus')
        )
        os.remove(os.path.join(self.download_dir, 'title.fidus'))

        # EPUB
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, '.dt-bulk-dropdown'))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Export selected as Epub"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(
            os.path.join(self.download_dir, 'title.epub')
        )
        os.remove(os.path.join(self.download_dir, 'title.epub'))

        # HTML
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, '.dt-bulk-dropdown'))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Export selected as HTML"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(
            os.path.join(self.download_dir, 'title.html.zip')
        )
        os.remove(os.path.join(self.download_dir, 'title.html.zip'))

        # LaTeX
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, '.dt-bulk-dropdown'))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Export selected as LaTeX"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(
            os.path.join(self.download_dir, 'title.latex.zip')
        )
        os.remove(os.path.join(self.download_dir, 'title.latex.zip'))

        # JATS
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, '.dt-bulk-dropdown'))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Export selected as JATS"]'
        ).click()
        time.sleep(1)
        assert os.path.isfile(
            os.path.join(self.download_dir, 'title.jats.zip')
        )
        os.remove(os.path.join(self.download_dir, 'title.jats.zip'))

        # Revision
        self.driver.find_element_by_css_selector(
            '.revisions'
        ).click()
        self.driver.find_element_by_css_selector(
            '.download-revision'
        ).click()
        time.sleep(1)
        assert os.path.isfile(
            os.path.join(self.download_dir, 'title.fidus')
        )
        # We keep the file to test import below
        self.driver.find_element_by_css_selector(
            '.recreate-revision'
        ).click()
        self.assertEqual(
            len(self.driver.find_elements_by_css_selector(
                '#revisions-dialog > table > tbody tr'
            )),
            1
        )
        self.driver.find_element_by_css_selector(
            '.delete-revision'
        ).click()
        self.driver.find_element_by_css_selector(
            'button.fw-dark'
        ).click()
        time.sleep(1)
        self.assertEqual(
            len(self.driver.find_elements_by_css_selector(
                '#revisions-dialog > table > tbody tr'
            )),
            0
        )
        self.driver.find_element_by_css_selector(
            'button.fw-orange'
        ).click()

        # Delete document
        documents = self.driver.find_elements_by_css_selector(
            '.fw-contents tbody tr a.doc-title'
        )
        self.assertEqual(
            len(documents),
            2
        )
        self.driver.find_element_by_css_selector(
            'tr:nth-child(1) > td > label'
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, '.dt-bulk-dropdown'))
        ).click()
        self.driver.find_element_by_xpath(
            '//*[normalize-space()="Delete selected"]'
        ).click()
        self.driver.find_element_by_css_selector(
            'button.fw-dark'
        ).click()
        time.sleep(1)
        documents = self.driver.find_elements_by_css_selector(
            '.fw-contents tbody tr a.doc-title'
        )
        self.assertEqual(
            len(documents),
            1
        )
        self.driver.find_element_by_css_selector(
            '.fw-contents tbody tr .delete-document'
        ).click()
        self.driver.find_element_by_css_selector(
            'button.fw-dark'
        ).click()
        time.sleep(1)
        documents = self.driver.find_elements_by_css_selector(
            '.fw-contents tbody tr a.doc-title'
        )
        self.assertEqual(
            len(documents),
            0
        )
        # We delete the image so our import will cause an image import
        self.driver.find_element_by_css_selector(
            "a[href='/usermedia/']"
        ).click()
        self.driver.find_element_by_css_selector(
            ".delete-image"
        ).click()
        self.driver.find_element_by_css_selector(
            ".fw-dark"
        ).click()
        # We import the fidus file
        self.driver.find_element_by_css_selector(
            "a[href='/']"
        ).click()
        self.driver.find_element_by_css_selector(
            "button[title='Upload Fidus document']"
        ).click()
        self.driver.find_element_by_css_selector(
            "#fidus-uploader"
        ).send_keys(os.path.join(self.download_dir, 'title.fidus'))
        self.driver.find_element_by_css_selector(
            ".fw-dark"
        ).click()
        documents = self.driver.find_elements_by_css_selector(
            '.fw-contents tbody tr a.doc-title'
        )
        self.assertEqual(
            len(documents),
            1
        )
        # We delete our downloaded fidus file
        os.remove(os.path.join(self.download_dir, 'title.fidus'))



import time
import json
from builtins import range

from django.http import JsonResponse
from django.contrib.auth.decorators import login_required
from django.db.models import Max, Count
from django.core.serializers.python import Serializer
from django.views.decorators.http import require_POST

from base.decorators import ajax_required

from bibliography.models import (
    Entry,
    EntryCategory
)


class SimpleSerializer(Serializer):

    def end_object(self, obj):
        self._current['id'] = obj._get_pk_val()
        self.objects.append(self._current)


serializer = SimpleSerializer()


# returns list of bibliography items
@login_required
@require_POST
@ajax_required
def biblist(request):
    response = {}
    last_modified_onclient = int(request.POST['last_modified'])
    number_of_entries_onclient = int(
        request.POST['number_of_entries'])
    aggregation_values = Entry.objects.filter(
        entry_owner=request.user.id).aggregate(
        Max('last_modified'), Count('id'))
    last_modified__max = aggregation_values[
        'last_modified__max']
    number_of_entries_onserver = aggregation_values[
        'id__count']
    if last_modified__max:
        last_modified_onserver = int(
            time.mktime(last_modified__max.timetuple()))
    else:
        last_modified_onserver = 0
    if (
        last_modified_onclient < last_modified_onserver or
        number_of_entries_onclient > number_of_entries_onserver or
        request.user.id != int(request.POST['user_id'])
    ):
        response['bib_list'] = serializer.serialize(
            Entry.objects.filter(
                entry_owner=request.user), fields=(
                    'entry_key',
                    'entry_owner',
                    'bib_type',
                    'entry_cat',
                    'fields'
                )
            )
        response['last_modified'] = last_modified_onserver
        response[
            'number_of_entries'] = number_of_entries_onserver
    response['bib_categories'] = serializer.serialize(
        EntryCategory.objects.filter(category_owner=request.user))
    response['user_id'] = request.user.id
    status = 200
    return JsonResponse(
        response,
        status=status
    )


# save bibliography entries from bibtex importer or form
@login_required
@require_POST
@ajax_required
def save(request):
    response = {}
    bibs = json.loads(request.POST['bibs'])
    status = 200
    response['id_translations'] = []
    for b_id in list(bibs.keys()):
        bib = bibs[b_id]
        if request.POST['is_new'] == 'true':
            inserting_obj = {
                'entry_owner_id': request.user.id,
                'entry_key': bib['entry_key'][-64:],
                'bib_type': bib['bib_type'],
                'entry_cat': bib['entry_cat'],
                'fields': bib['fields']
            }
            similar = Entry.objects.filter(**inserting_obj).first()
            if similar:
                response['id_translations'].append([b_id, similar.id])
            else:
                the_entry = Entry(**inserting_obj)
                the_entry.save()
                response['id_translations'].append([b_id, the_entry.id])

        else:
            the_entry = Entry.objects.get(id=b_id)
            the_entry.entry_key = bib['entry_key'][-64:]
            the_entry.bib_type = bib['bib_type']
            the_entry.entry_cat = bib['entry_cat']
            the_entry.fields = bib['fields']
            the_entry.save()
            response['id_translations'].append([b_id, the_entry.id])
    return JsonResponse(
        response,
        status=status
    )


# delete an entry
@login_required
@require_POST
@ajax_required
def delete(request):
    response = {}
    status = 201
    ids = request.POST.getlist('ids[]')
    id_chunks = [ids[x:x + 100] for x in range(0, len(ids), 100)]
    for id_chunk in id_chunks:
        Entry.objects.filter(
            pk__in=id_chunk,
            entry_owner=request.user
        ).delete()
    return JsonResponse(
        response,
        status=status
    )


# save changes or create a new category
@login_required
@require_POST
@ajax_required
def save_category(request):
    response = {}
    response['entries'] = []
    ids = request.POST.getlist('ids[]')
    titles = request.POST.getlist('titles[]')
    EntryCategory.objects.filter(
        category_owner=request.user
    ).exclude(id__in=ids).delete()
    x = 0
    for the_id in ids:
        the_id = int(the_id)
        the_title = titles[x]
        x += 1
        if 0 == the_id:
            # if the category is new, then create new
            the_cat = EntryCategory(
                category_title=the_title,
                category_owner=request.user
            )
        else:
            # if the category already exists, update the title
            the_cat = EntryCategory.objects.get(pk=the_id)
            the_cat.category_title = the_title
        the_cat.save()
        response['entries'].append(
            {'id': the_cat.id, 'category_title': the_cat.category_title}
        )
    status = 201

    return JsonResponse(
        response,
        status=status
    )


# Delete one or more categories
@login_required
@require_POST
@ajax_required
def delete_category(request):
    response = {}
    status = 201
    ids = request.POST.getlist('ids[]')
    EntryCategory.objects.filter(
        pk__in=ids,
        entry_owner=request.user
    ).delete()
    return JsonResponse(
        response,
        status=status
    )



import json

from django.http import JsonResponse
from django.contrib.auth import logout, update_session_auth_hash
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.conf import settings
from django.shortcuts import HttpResponseRedirect
from django.views.decorators.http import require_POST

from base.decorators import ajax_required
from .forms import UserForm, TeamMemberForm
from . import util as userutil
from document.models import AccessRight, AccessRightInvite

from allauth.account.models import (
    EmailAddress,
    EmailConfirmation,
    EmailConfirmationHMAC
)
from allauth.account.views import SignupView
from allauth.account import signals
from allauth.socialaccount.models import SocialAccount
from allauth.socialaccount.signals import social_account_removed
from django.contrib.auth.forms import PasswordChangeForm
from allauth.account.forms import AddEmailForm

from avatar.models import Avatar
from avatar import views as avatarviews
from avatar.forms import UploadAvatarForm
from avatar.signals import avatar_updated

from document.views import apply_invite


# Outdated but we need it to allow for update of 3.7 instances.
# Can be removed in 3.9.
def info(request):
    return JsonResponse({'is_authenticated': False})


@login_required
@ajax_required
@require_POST
def password_change(request):
    '''
    Change password
    '''
    response = {}
    form = PasswordChangeForm(user=request.user, data=request.POST)
    if form.is_valid():
        status = 200
        form.save()
        # Updating the password logs out all other sessions for the user
        # except the current one.
        update_session_auth_hash(request, form.user)
    else:
        response['msg'] = form.errors
        status = 201

    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def add_email(request):
    '''
    Add email address
    '''
    response = {}
    add_email_form = AddEmailForm(request.user, request.POST)
    if add_email_form.is_valid():
        status = 200
        email_address = add_email_form.save(request)
        signals.email_added.send(
            sender=request.user.__class__,
            request=request, user=request.user,
            email_address=email_address
        )
    else:
        status = 201
        response['msg'] = add_email_form.errors

    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def delete_email(request):
    response = {}
    email = request.POST["email"]
    response['msg'] = "Removed e-mail address " + email
    status = 200
    email_address = EmailAddress.objects.filter(
        user=request.user,
        email=email,
        primary=False
    ).first()
    if not email_address:
        return JsonResponse(
            {'msg': 'Cannot remove email.'},
            status=404
        )
    email_address.delete()
    signals.email_removed.send(
        sender=request.user.__class__,
        request=request,
        user=request.user,
        email_address=email_address
    )
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def primary_email(request):
    response = {}
    email = request.POST["email"]
    email_address = EmailAddress.objects.filter(
        user=request.user,
        email=email,
        verified=True
    ).first()
    if not email_address:
        return JsonResponse({'msg': 'Cannot set primary email.'}, 404)
    from_email_address = EmailAddress.objects.filter(
        user=request.user,
        primary=True
    ).first()
    status = 200
    email_address.set_as_primary()
    response['msg'] = "Primary e-mail address set"
    signals.email_changed.send(
        sender=request.user.__class__,
        request=request, user=request.user,
        from_email_address=from_email_address,
        to_email_address=email_address
    )
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def delete_socialaccount(request):
    account = SocialAccount.objects.filter(
        id=request.POST["socialaccount"],
        user=request.user
    ).first()
    if not account:
        return JsonResponse(
            {'msg': 'Unknown account'},
            status=404
        )
    account.delete()
    social_account_removed.send(
        sender=SocialAccount,
        request=request,
        socialaccount=account
    )
    return JsonResponse(
        {'msg': 'Deleted account'},
        status=200
    )


@login_required
@ajax_required
@require_POST
def upload_avatar(request):
    '''
    Upload avatar image
    '''
    response = {}
    status = 405

    avatar, avatars = avatarviews._get_avatars(request.user)
    upload_avatar_form = UploadAvatarForm(
        None,
        request.FILES,
        user=request.user
    )
    if upload_avatar_form.is_valid():
        avatar = Avatar(
            user=request.user,
            primary=True,
        )
        image_file = request.FILES['avatar']
        avatar.avatar.save(image_file.name, image_file)
        avatar.save()
        avatar_updated.send(
            sender=Avatar,
            user=request.user,
            avatar=avatar
        )
        response['avatar'] = userutil.get_user_avatar_url(
            request.user
        )['url']
        status = 200
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def delete_avatar(request):
    '''
    Delete avatar image
    '''
    response = {}
    status = 405
    avatar, avatars = avatarviews._get_avatars(request.user)
    if avatar is None:
        response['error'] = 'User has no avatar'
    else:
        aid = avatar.id
        for a in avatars:
            if a.id == aid:
                a.primary = True
                a.save()
                avatar_updated.send(
                    sender=Avatar,
                    user=request.user,
                    avatar=avatar
                )
                break
        Avatar.objects.filter(pk=aid).delete()
        response['avatar'] = userutil.get_user_avatar_url(
            request.user
        )['url']
        status = 200
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def delete_user(request):
    """
    Delete the user
    """
    response = {}
    user = request.user
    # Only remove users who are not marked as having staff status
    # to prevent administratoras from deleting themselves accidentally.
    if not user.check_password(request.POST['password']):
        status = 401
    elif user.is_staff:
        status = 403
    else:
        logout(request)
        user.delete()
        status = 204
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def save_profile(request):
    """
    Save user profile information
    """
    response = {}
    form_data = json.loads(request.POST['form_data'])
    user_object = User.objects.get(pk=request.user.pk)
    user_form = UserForm(form_data['user'], instance=user_object)
    if user_form.is_valid():
        user_form.save()
        status = 200
    else:
        response['errors'] = user_form.errors
        status = 422
    '''
    currently not used
    profile_object = user_object.profile
    profile_form = UserProfileForm(
        form_data['profile'],
        instance=profile_object
    )
    if profile_form.is_valid():
        if status == 200:
            user_form.save()
            profile_form.save()
    else:
        if status == 200:
            response['errors']=profile_form.errors
            status = 422
        else:
            response['errors']+=profile_form.errors
    '''

    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def list_team_members(request):
    response = {}
    status = 200
    response['team_members'] = []

    for member in User.objects.filter(member__leader=request.user):
        team_member = {
            'id': member.id,
            'name': member.readable_name,
            'username': member.get_username(),
            'email': member.email,
            'avatar': userutil.get_user_avatar_url(member)
        }
        response['team_members'].append(team_member)
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def add_team_member(request):
    """
    Add a user as a team member of the current user
    """
    response = {}
    new_member = False
    status = 202
    user_string = request.POST['user_string']
    if "@" in user_string and "." in user_string:
        email_address = EmailAddress.objects.filter(
            email=user_string
        ).first()
        if email_address:
            new_member = email_address.user
    else:
        user = User.objects.filter(username=user_string).first()
        if user:
            new_member = user
    if new_member:
        if new_member.pk is request.user.pk:
            # 'You cannot add yourself to your contacts!'
            response['error'] = 1
        else:
            form_data = {
                'leader': request.user.pk,
                'member': new_member.pk
            }
            team_member_form = TeamMemberForm(form_data)
            if team_member_form.is_valid():
                team_member_form.save()
                the_avatar = userutil.get_user_avatar_url(new_member)
                response['member'] = {
                    'id': new_member.pk,
                    'name': new_member.username,
                    'email': new_member.email,
                    'avatar': the_avatar
                }
                status = 201
            else:
                # 'This person is already in your contacts!'
                response['error'] = 2
    else:
        # 'User cannot be found'
        response['error'] = 3

    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def remove_team_member(request):
    """
    Remove a team member
    """
    response = {}
    former_members = request.POST.getlist('members[]')
    for former_member in former_members:
        former_member = int(former_member)
        # Revoke all permissions given to this person
        AccessRight.objects.filter(
            user_id=former_member,
            document__owner=request.user
        ).delete()
        # Now delete the user from the team
        team_member_object_instance = request.user.leader.filter(
            member_id=former_member
        ).first()
        team_member_object_instance.delete()
    status = 200
    return JsonResponse(
        response,
        status=status
    )


@ajax_required
@require_POST
def get_confirmkey_data(request):
    """
    Get data for an email confirmation key
    """
    response = {}
    key = request.POST['key']
    confirmation = EmailConfirmationHMAC.from_key(key)
    if not confirmation:
        qs = EmailConfirmation.objects.all_valid()
        qs = qs.select_related("email_address__user")
        confirmation = qs.filter(key=key.lower()).first()
    if confirmation:
        status = 200
        response['username'] = confirmation.email_address.user.username
        response['email'] = confirmation.email_address.email
        if request.user:
            if request.user != confirmation.email_address.user:
                response['logout'] = True
                logout(request)
        # We check if the user has another verified email already. If yes,
        # we don't need to display the terms and test server warning again.
        if confirmation.email_address.user.emailaddress_set.filter(
            verified=True
        ).first():
            response['verified'] = True
        else:
            response['verified'] = False
    else:
        status = 404
    return JsonResponse(
        response,
        status=status
    )


class FidusSignupView(SignupView):
    def form_valid(self, form):
        if not settings.REGISTRATION_OPEN:
            return HttpResponseRedirect('/')
        ret = super().form_valid(form)
        if ret.status_code > 399:
            return ret
        if 'invite_id' in self.request.POST:
            invite_id = int(self.request.POST['invite_id'])
            inv = AccessRightInvite.objects.filter(id=invite_id).first()
            if inv:
                apply_invite(inv, self.user)
        else:
            inv = AccessRightInvite.objects.filter(
                email=self.user.email
            ).first()
            if inv:
                apply_invite(inv, self.user)
        return ret


signup = FidusSignupView.as_view()



import time

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.action_chains import ActionChains

from testing.testcases import LiveTornadoTestCase
from testing.selenium_helper import SeleniumHelper


class ContactsTest(LiveTornadoTestCase, SeleniumHelper):
    fixtures = [
        'initial_documenttemplates.json',
        'initial_styles.json',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.base_url = cls.live_server_url
        driver_data = cls.get_drivers(1)
        cls.driver = driver_data["drivers"][0]
        cls.client = driver_data["clients"][0]
        cls.driver.implicitly_wait(driver_data["wait_time"])
        cls.wait_time = driver_data["wait_time"]

    @classmethod
    def tearDownClass(cls):
        cls.driver.quit()
        super().tearDownClass()

    def setUp(self):
        self.user = self.create_user(
            username='Yeti',
            email='yeti@snowman.com',
            passtext='password'
        )
        self.contact1 = self.create_user(
            username='Contact1',
            email='contact1@snowman.com',
            passtext='password'
        )
        self.contact2 = self.create_user(
            username='Contact2',
            email='contact2@snowman.com',
            passtext='password'
        )
        self.login_user(self.user, self.driver, self.client)

    def test_contacts(self):
        self.driver.get(self.base_url + "/")
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    ".new_document button"
                )
            )
        ).click()
        WebDriverWait(self.driver, self.wait_time).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'editor-toolbar'))
        )
        # Open share dialog
        self.driver.find_element(
            By.CSS_SELECTOR,
            ".header-menu:nth-child(1) > .header-nav-item"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            "li:nth-child(1) > .fw-pulldown-item"
        ).click()
        self.driver.find_element(
            By.CSS_SELECTOR,
            ".ui-dialog .fw-add-button"
        ).click()
        self.driver.find_element(
            By.ID,
            "new-member-user-string"
        ).click()
        self.driver.find_element(By.ID, "new-member-user-string").send_keys(
            "contact1@snowman.com"
        )
        ActionChains(self.driver).send_keys(
            Keys.TAB
        ).send_keys(
            Keys.RETURN
        ).perform()
        self.driver.find_element(
            By.CSS_SELECTOR,
            ".ui-dialog .fw-dark"
        ).click()
        # Close document
        self.driver.find_element(
            By.ID,
            "close-document-top"
        ).click()
        # Enter contacts page
        WebDriverWait(self.driver, self.wait_time).until(
            EC.element_to_be_clickable((By.ID, 'preferences-btn'))
        ).click()
        self.driver.find_element_by_css_selector(
            "a[href='/user/team/']"
        ).click()
        self.assertEqual(
            len(self.driver.find_elements_by_css_selector(
                '#team-table .entry-select'
            )),
            1
        )
        self.driver.find_element_by_css_selector(
            "button[title='Add new contact']"
        ).click()
        self.driver.find_element_by_id(
            "new-member-user-string"
        ).send_keys('contact2@snowman.com')
        self.driver.find_element_by_css_selector(
            "button.fw-dark"
        ).click()
        time.sleep(1)
        self.assertEqual(
            len(self.driver.find_elements_by_css_selector(
                '#team-table .entry-select'
            )),
            2
        )
        # Delete individual
        self.driver.find_element_by_css_selector(
            ".delete-single-member"
        ).click()
        self.driver.find_element_by_css_selector(
            "button.fw-dark"
        ).click()
        time.sleep(1)
        self.assertEqual(
            len(self.driver.find_elements_by_css_selector(
                '#team-table .entry-select'
            )),
            1
        )
        self.driver.find_element_by_css_selector(
            "#team-table td:nth-child(1) > label"
        ).click()
        self.driver.find_element_by_css_selector(
            ".dt-bulk-dropdown"
        ).click()
        self.driver.find_element_by_css_selector(
            "li[title='Delete selected contacts.']"
        ).click()
        self.driver.find_element_by_css_selector(
            "button.fw-dark"
        ).click()
        time.sleep(1)
        self.assertEqual(
            len(self.driver.find_elements_by_css_selector(
                '#team-table .entry-select'
            )),
            0
        )



import time
import os
import bleach
import json
from tornado.escape import json_decode, json_encode
from django.core import serializers
from django.http import HttpResponse, JsonResponse, HttpRequest
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_POST
from django.db import transaction
from django.core.files import File
from django.contrib.auth.models import User
from django.utils.translation import ugettext as _
from django.conf import settings
from django.core.mail import send_mail
from django.db.models import F, Q
from django.contrib.admin.views.decorators import staff_member_required

from user.util import get_user_avatar_url
from document.models import Document, AccessRight, DocumentRevision, \
    DocumentTemplate, AccessRightInvite, CAN_UPDATE_DOCUMENT, \
    CAN_COMMUNICATE, FW_DOCUMENT_VERSION
from usermedia.models import DocumentImage, Image
from bibliography.models import Entry
from document.helpers.serializers import PythonWithURLSerializer
from bibliography.views import serializer
from style.models import DocumentStyle
from base.html_email import html_email
from base.decorators import ajax_required
from user.models import TeamMember


@login_required
@ajax_required
@require_POST
def get_documentlist_extra(request):
    response = {}
    status = 200
    ids = request.POST['ids'].split(',')
    docs = Document.objects.filter(Q(owner=request.user) | Q(
        accessright__user=request.user)).filter(id__in=ids)
    response['documents'] = []
    for doc in docs:
        images = {}
        for image in doc.documentimage_set.all():
            images[image.image.id] = {
                'added': image.image.added,
                'checksum': image.image.checksum,
                'file_type': image.image.file_type,
                'height': image.image.height,
                'id': image.image.id,
                'image': image.image.image.url,
                'thumbnail': image.image.thumbnail.url,
                'title': image.title,
                'copyright': json.loads(image.copyright),
                'width': image.image.width
            }
        response['documents'].append({
            'images': images,
            'contents': doc.contents,
            'comments': doc.comments,
            'bibliography': doc.bibliography,
            'id': doc.id
        })
    return JsonResponse(
        response,
        status=status
    )


@login_required
def documents_list(request):
    documents = Document.objects.filter(
        Q(owner=request.user) | Q(accessright__user=request.user),
        listed=True
    ).order_by('-updated')
    output_list = []
    for document in documents:
        if document.owner == request.user:
            access_right = 'write'
        else:
            access_right = AccessRight.objects.get(
                user=request.user,
                document=document
            ).rights
        if (
            request.user.is_staff or
            document.owner == request.user or
            AccessRight.objects.filter(
                document=document,
                user=request.user,
                rights__in=CAN_COMMUNICATE
            ).first()
        ):
            revisions = DocumentRevision.objects.filter(document=document)
            revision_list = []
            for revision in revisions:
                revision_list.append({
                    'date': time.mktime(revision.date.utctimetuple()),
                    'note': revision.note,
                    'file_name': revision.file_name,
                    'pk': revision.pk
                })
        else:
            revision_list = []
        added = time.mktime(document.added.utctimetuple())
        updated = time.mktime(document.updated.utctimetuple())
        is_owner = False
        if document.owner == request.user:
            is_owner = True
        output_list.append({
            'id': document.id,
            'title': document.title,
            'is_owner': is_owner,
            'owner': {
                'id': document.owner.id,
                'name': document.owner.readable_name,
                'avatar': get_user_avatar_url(document.owner)
            },
            'added': added,
            'updated': updated,
            'rights': access_right,
            'revisions': revision_list
        })
    return output_list


@login_required
@ajax_required
@require_POST
def get_access_rights(request):
    response = {}
    status = 200
    ar_qs = AccessRight.objects.filter(document__owner=request.user)
    in_qs = AccessRightInvite.objects.filter(document__owner=request.user)
    doc_ids = request.POST.getlist('document_ids[]')
    if len(doc_ids) > 0:
        ar_qs = ar_qs.filter(document_id__in=doc_ids)
        in_qs = in_qs.filter(document_id__in=doc_ids)
    access_rights = []
    for ar in ar_qs:
        access_rights.append({
            'document_id': ar.document.id,
            'user_id': ar.user.id,
            'user_name': ar.user.readable_name,
            'rights': ar.rights,
            'avatar': get_user_avatar_url(ar.user)
        })
    response['access_rights'] = access_rights
    invites = []
    for inv in in_qs:
        invites.append({
            'document_id': inv.document.id,
            'email': inv.email,
            'rights': inv.rights
        })
    response['invites'] = invites
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
@transaction.atomic
def save_access_rights(request):
    response = {}
    doc_ids = json_decode(request.POST['document_ids'])
    rights = json_decode(request.POST['access_rights'])
    invites = json_decode(request.POST['invites'])
    for doc_id in doc_ids:
        doc = Document.objects.filter(
            pk=doc_id,
            owner=request.user
        ).first()
        if not doc:
            continue
        for right in rights:
            if right['rights'] == 'delete':
                # Status 'delete' means the access right is marked for
                # deletion.
                AccessRight.objects.filter(
                    document_id=doc_id,
                    user_id=right['user_id']
                ).delete()
            else:
                access_right = AccessRight.objects.filter(
                    document_id=doc_id,
                    user_id=right['user_id']
                ).first()
                if access_right:
                    if access_right.rights != right['rights']:
                        access_right.rights = right['rights']
                        send_share_notification(
                            request,
                            doc_id,
                            right['user_id'],
                            right['rights'],
                            True
                        )
                else:
                    access_right = AccessRight.objects.create(
                        document_id=doc_id,
                        user_id=right['user_id'],
                        rights=right['rights']
                    )
                    send_share_notification(
                        request,
                        doc_id,
                        right['user_id'],
                        right['rights'],
                        False
                    )
                access_right.save()
        for invite in invites:
            if invite['rights'] == 'delete':
                # Status 'delete' means the invite is marked for
                # deletion.
                AccessRightInvite.objects.filter(
                    document_id=doc_id,
                    email=invite['email']
                ).delete()
            else:
                old_invite = AccessRightInvite.objects.filter(
                    document_id=doc_id,
                    email=invite['email']
                ).first()
                if old_invite:
                    if old_invite.rights != invite['rights']:
                        old_invite.rights = invite['rights']
                        old_invite.save()
                        send_invite_notification(
                            request,
                            doc_id,
                            invite['email'],
                            invite['rights'],
                            old_invite,
                            True
                        )
                else:
                    new_invite = AccessRightInvite.objects.create(
                        document_id=doc_id,
                        email=invite['email'],
                        rights=invite['rights']
                    )
                    new_invite.save()
                    send_invite_notification(
                        request,
                        doc_id,
                        invite['email'],
                        invite['rights'],
                        new_invite,
                        False
                    )
    status = 201
    return JsonResponse(
        response,
        status=status
    )


def apply_invite(inv, user):
    old_ar = AccessRight.objects.filter(
        user=user,
        document=inv.document
    ).first()
    if old_ar:
        # If the user already has rights, we should only be upgrading
        # them, not downgrade. Unfortuantely it is not easy to
        # say how each right compares. So unless the invite gives read
        # access, or the user already has write access, we change to
        # the access right of the invite.
        if inv.rights == 'read':
            pass
        elif old_ar.rights == 'write':
            pass
        else:
            old_ar.rights = inv.rights
            old_ar.save()
    elif inv.document.owner == user:
        pass
    else:
        ar = AccessRight.objects.create(
            document=inv.document,
            user=user,
            rights=inv.rights
        )
        ar.save()
        if not TeamMember.objects.filter(
            leader=inv.document.owner,
            member=user
        ).first():
            tm1 = TeamMember.objects.create(
                leader=inv.document.owner,
                member=user
            )
            tm1.save()
        if not TeamMember.objects.filter(
            leader=user,
            member=inv.document.owner,
        ).first():
            tm2 = TeamMember.objects.create(
                leader=user,
                member=inv.document.owner
            )
            tm2.save()
    inv.delete()


@login_required
@ajax_required
@require_POST
def invite(request):
    response = {}
    status = 200
    id = int(request.POST['id'])
    inv = AccessRightInvite.objects.filter(id=id).first()
    if inv:
        response['redirect'] = inv.document.get_absolute_url()
        apply_invite(inv, request.user)
    else:
        response['redirect'] = ''
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def get_documentlist(request):
    response = {}
    status = 200
    response['documents'] = documents_list(request)
    response['team_members'] = []
    for team_member in request.user.leader.all():
        tm_object = {}
        tm_object['id'] = team_member.member.id
        tm_object['name'] = team_member.member.readable_name
        tm_object['username'] = team_member.member.get_username()
        tm_object['avatar'] = get_user_avatar_url(team_member.member)
        response['team_members'].append(tm_object)
    serializer = PythonWithURLSerializer()
    doc_styles = serializer.serialize(
        DocumentStyle.objects.filter(
            Q(document_template__user=None) |
            Q(document_template__user=request.user)
        ),
        use_natural_foreign_keys=True,
        fields=['title', 'slug', 'contents', 'documentstylefile_set']
    )
    response['document_styles'] = [obj['fields'] for obj in doc_styles]
    doc_templates = DocumentTemplate.objects.filter(
        Q(user=request.user) | Q(user=None)
    ).order_by(F('user').desc(nulls_first=True))
    response['document_templates'] = {}
    for obj in doc_templates:
        response['document_templates'][obj.import_id] = {
            'title': obj.title,
            'id': obj.id
        }
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def delete(request):
    response = {}
    status = 200
    doc_id = int(request.POST['id'])
    document = Document.objects.get(pk=doc_id, owner=request.user)
    if document.is_deletable():
        image_ids = list(
            DocumentImage.objects.filter(document_id=doc_id)
            .values_list('image_id', flat=True)
        )
        document.delete()
        for image in Image.objects.filter(id__in=image_ids):
            if image.is_deletable():
                image.delete()
        response['done'] = True
    else:
        response['done'] = False
    return JsonResponse(
        response,
        status=status
    )


def send_share_notification(request, doc_id, collaborator_id, rights, change):
    owner = request.user.readable_name
    document = Document.objects.get(id=doc_id)
    collaborator = User.objects.get(id=collaborator_id)
    collaborator_name = collaborator.readable_name
    collaborator_email = collaborator.email
    document_title = document.title
    if len(document_title) == 0:
        document_title = _('Untitled')
    link = HttpRequest.build_absolute_uri(request, document.get_absolute_url())
    if change:
        message_text = _(
            ('Hey %(collaborator_name)s,\n%(owner)s has changed your access '
             'rights to %(rights)s on the document \'%(document_title)s\'. '
             '\nAccess the document through this link: %(link)s')
        ) % {
            'owner': owner,
            'rights': rights,
            'collaborator_name': collaborator_name,
            'link': link,
            'document_title': document_title
        }
        body_html_intro = _(
            ('<p>Hey %(collaborator_name)s,<br>%(owner)s has changed your '
             'access rights to %(rights)s on the document '
             '\'%(document_title)s\'.</p>')
        ) % {
            'owner': owner,
            'rights': rights,
            'collaborator_name': collaborator_name,
            'document_title': document_title
        }
    else:
        message_text = _(
            ('Hey %(collaborator_name)s,\n%(owner)s has shared the document '
             '\'%(document_title)s\' with you and given you %(rights)s access '
             'rights. '
             '\nAccess the document through this link: %(link)s')
        ) % {
            'owner': owner,
            'rights': rights,
            'collaborator_name': collaborator_name,
            'link': link,
            'document_title': document_title
        }
        body_html_intro = _(
            ('<p>Hey %(collaborator_name)s,<br>%(owner)s has shared the '
             'document \'%(document_title)s\' with you and given you '
             '%(rights)s access rights.</p>')
        ) % {
            'owner': owner,
            'rights': rights,
            'collaborator_name': collaborator_name,
            'document_title': document_title
        }

    body_html = (
        '<h1>%(document_title)s %(shared)s</h1>'
        '%(body_html_intro)s'
        '<table>'
        '<tr><td>'
        '%(Document)s'
        '</td><td>'
        '<b>%(document_title)s</b>'
        '</td></tr>'
        '<tr><td>'
        '%(Author)s'
        '</td><td>'
        '%(owner)s'
        '</td></tr>'
        '<tr><td>'
        '%(AccessRights)s'
        '</td><td>'
        '%(rights)s'
        '</td></tr>'
        '</table>'
        '<div class="actions"><a class="button" href="%(link)s">'
        '%(AccessTheDocument)s'
        '</a></div>'
    ) % {
        'shared': _('shared'),
        'body_html_intro': body_html_intro,
        'Document': _('Document'),
        'document_title': document_title,
        'Author': _('Author'),
        'owner': owner,
        'AccessRights': _('Access Rights'),
        'rights': rights,
        'link': link,
        'AccessTheDocument': _('Access the document')
    }
    send_mail(
        _('Document shared:') +
        ' ' +
        document_title,
        message_text,
        settings.DEFAULT_FROM_EMAIL,
        [collaborator_email],
        fail_silently=True,
        html_message=html_email(body_html)
    )


def send_invite_notification(request, doc_id, email, rights, invite, change):
    owner = request.user.readable_name
    document = Document.objects.get(id=doc_id)
    document_title = document.title
    if len(document_title) == 0:
        document_title = _('Untitled')
    link = HttpRequest.build_absolute_uri(request, invite.get_absolute_url())
    if change:
        message_text = _(
            ('Hey %(email)s,\nas we told you previously, %(owner)s has '
             'invited you to join Fidus Writer and shared the document '
             '\'%(document_title)s\' with you. '
             '\n%(owner)s has now changed your access rights to %(rights)s.'
             '\nAccess the document through this link: %(link)s')
        ) % {
            'owner': owner,
            'rights': rights,
            'email': email,
            'link': link,
            'document_title': document_title
        }
        body_html_intro = _(
            ('<p>Hey %(email)s,<br>as we told you previously, '
             '%(owner)s has invited you to join Fidus Writer and shared the '
             '\'%(document_title)s\' with you.</p>'
             '<p>%(owner)s has now changed your access rights to %(rights)s. '
             '</p>')
        ) % {
            'owner': owner,
            'rights': rights,
            'email': email,
            'document_title': document_title
        }
    else:
        message_text = _(
            ('Hey %(email)s,\n%(owner)s has invited you to Fidus '
             ' Writer, shared the document \'%(document_title)s\' with you, '
             'and given you %(rights)s access rights. '
             '\nAccess the document through this link: %(link)s')
        ) % {
            'owner': owner,
            'rights': rights,
            'email': email,
            'link': link,
            'document_title': document_title
        }
        body_html_intro = _(
            ('<p>Hey %(email)s,<br>%(owner)s has invited you to '
             'Fidus Writer, shared the document \'%(document_title)s\' with '
             'you, and given you %(rights)s access rights.</p>')
        ) % {
            'owner': owner,
            'rights': rights,
            'email': email,
            'document_title': document_title
        }

    body_html = (
        '<h1>%(document_title)s %(shared)s</h1>'
        '%(body_html_intro)s'
        '<table>'
        '<tr><td>'
        '%(Document)s'
        '</td><td>'
        '<b>%(document_title)s</b>'
        '</td></tr>'
        '<tr><td>'
        '%(Author)s'
        '</td><td>'
        '%(owner)s'
        '</td></tr>'
        '<tr><td>'
        '%(AccessRights)s'
        '</td><td>'
        '%(rights)s'
        '</td></tr>'
        '</table>'
        '<div class="actions"><a class="button" href="%(link)s">'
        '%(AccessTheDocument)s'
        '</a></div>'
    ) % {
        'shared': _('shared'),
        'body_html_intro': body_html_intro,
        'Document': _('Document'),
        'document_title': document_title,
        'Author': _('Author'),
        'owner': owner,
        'AccessRights': _('Access Rights'),
        'rights': rights,
        'link': link,
        'AccessTheDocument': _('Sign up or log in and access the document')
    }
    send_mail(
        _('Document shared:') +
        ' ' +
        document_title,
        message_text,
        settings.DEFAULT_FROM_EMAIL,
        [email],
        fail_silently=True,
        html_message=html_email(body_html)
    )


@login_required
@ajax_required
@require_POST
def create_doc(request, template_id):
    response = {}
    document_template = DocumentTemplate.objects.filter(
        Q(user=request.user) | Q(user=None),
        id=template_id
    ).first()
    if not document_template:
        return JsonResponse(
            response,
            status=405
        )
    document = Document.objects.create(
        owner_id=request.user.pk,
        template_id=template_id
    )
    response['id'] = document.id
    return JsonResponse(
        response,
        status=201
    )


@login_required
@ajax_required
@require_POST
def import_create(request):
    # First step of import: Create a document and return the id of it
    response = {}
    status = 201
    import_id = request.POST['import_id']
    document_template = DocumentTemplate.objects.filter(
        Q(user=request.user) | Q(user=None),
        import_id=import_id
    ).order_by(F('user').desc(nulls_last=True)).first()
    if not document_template:
        # The user doesn't have this template.
        # We check whether the template exists with one of the documents
        # shared with the user. If so, we'll copy it so that we can avoid
        # having to create an entirely new template without styles or
        # exporter templates
        access_right = request.user.accessright_set.filter(
            document__template__import_id=import_id
        ).first()
        if access_right:
            document_template = access_right.document.template
            document_styles = list(
                document_template.documentstyle_set.all()
            )
            export_templates = list(
                document_template.exporttemplate_set.all()
            )
            document_template.pk = None
            document_template.user = request.user
            document_template.save()
            for ds in document_styles:
                style_files = list(ds.documentstylefile_set.all())
                ds.pk = None
                ds.document_template = document_template
                ds.save()
                for sf in style_files:
                    sf.pk = None
                    sf.style = ds
                    sf.save()
            for et in export_templates:
                et.pk = None
                et.document_template = document_template
                et.save()
    if not document_template:
        title = request.POST['template_title']
        definition = json_encode(json_decode(request.POST['template']))
        document_template = DocumentTemplate()
        document_template.title = title
        document_template.import_id = import_id
        document_template.user = request.user
        document_template.definition = definition
        document_template.save()
    document = Document.objects.create(
        owner=request.user,
        template=document_template
    )
    response['id'] = document.id
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def import_image(request):
    # create an image for a document
    response = {}
    document = Document.objects.filter(
        owner_id=request.user.pk,
        id=int(request.POST['doc_id'])
    ).first()
    if document:
        status = 201
    else:
        status = 401
        return JsonResponse(
            response,
            status=status
        )
    checksum = request.POST['checksum']
    image = Image.objects.filter(checksum=checksum).first()
    if image is None:
        image = Image.objects.create(
            uploader=request.user,
            image=request.FILES['image'],
            checksum=checksum
        )
    doc_image = DocumentImage.objects.create(
        image=image,
        title=request.POST['title'],
        copyright=json.dumps(request.POST['copyright']),
        document=document
    )
    response['id'] = doc_image.image.id
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def import_doc(request):
    response = {}
    doc_id = request.POST['id']
    # There is a doc_id, so we overwrite an existing doc rather than
    # creating a new one.
    document = Document.objects.get(id=int(doc_id))
    if (
        document.owner != request.user and not
        AccessRight.objects.filter(
            document_id=doc_id,
            user_id=request.user.id,
            rights__in=CAN_UPDATE_DOCUMENT
        ).first()
    ):
        response['error'] = 'No access to file'
        status = 403
        return JsonResponse(
            response,
            status=status
        )
    document.title = request.POST['title']
    # We need to decode/encode the following so that it has the same
    # character encoding as used the the save_document method in ws_views.
    document.contents = json_encode(json_decode(request.POST['contents']))
    document.comments = json_encode(json_decode(request.POST['comments']))
    document.bibliography = \
        json_encode(json_decode(request.POST['bibliography']))
    # document.doc_version should always be the current version, so don't
    # bother about it.
    document.save()
    response['document_id'] = document.id
    response['added'] = time.mktime(document.added.utctimetuple())
    response['updated'] = time.mktime(document.updated.utctimetuple())
    status = 200
    return JsonResponse(
        response,
        status=status
    )


@login_required
@ajax_required
@require_POST
def upload_revision(request):
    response = {}
    status = 405
    can_save = False
    document_id = request.POST['document_id']
    document = Document.objects.filter(id=int(document_id)).first()
    if document:
        if document.owner == request.user:
            can_save = True
        else:
            access_rights = AccessRight.objects.filter(
                document=document, user=request.user)
            if len(access_rights) > 0 and access_rights[
                0
            ].rights == 'write':
                can_save = True
    if can_save:
        status = 201
        revision = DocumentRevision()
        revision.file_object = request.FILES['file']
        revision.file_name = request.FILES['file'].name
        revision.note = request.POST['note']
        revision.document_id = document_id
        revision.save()
    return JsonResponse(
        response,
        status=status
    )


# Download a revision that was previously uploaded
@login_required
def get_revision(request, revision_id):
    revision = DocumentRevision.objects.filter(pk=int(revision_id)).first()
    if revision and (
        request.user.is_staff or
        revision.document.owner == request.user or
        AccessRight.objects.filter(
            document=revision.document,
            user=request.user,
            rights__in=CAN_COMMUNICATE
        ).first()
    ):
        http_response = HttpResponse(
            revision.file_object.file,
            content_type='application/zip; charset=x-user-defined',
            status=200
        )
        http_response[
            'Content-Disposition'] = 'attachment; filename=some_name.zip'
    else:
        http_response = HttpResponse(status=404)
    return http_response


@login_required
@ajax_required
@require_POST
def delete_revision(request):
    response = {}
    status = 405
    revision_id = request.POST['id']
    revision = DocumentRevision.objects.filter(pk=int(revision_id)).first()
    if revision:
        document = revision.document
        if document.owner == request.user:
            status = 200
            revision.delete()
    return JsonResponse(
        response,
        status=status
    )


# Check doc access rights.
def has_doc_access(doc, user):
    if doc.owner == user:
        return True
    access_rights = AccessRight.objects.filter(
        document=doc,
        user=user
    ).first()
    if access_rights:
        return True
    else:
        return False


@login_required
@ajax_required
@require_POST
def comment_notify(request):
    response = {}
    doc_id = request.POST['doc_id']
    collaborator_id = request.POST['collaborator_id']
    comment_text = request.POST['comment_text']
    comment_html = bleach.clean(
        request.POST['comment_html'],
        strip=True
    )
    notification_type = request.POST['type']
    collaborator = User.objects.filter(pk=collaborator_id).first()
    document = Document.objects.filter(pk=doc_id).first()
    if (
        not document or
        not collaborator or
        not comment_text or
        not comment_html or
        not has_doc_access(document, request.user) or
        not notification_type
    ):
        return JsonResponse(
            response,
            status=403
        )
    if (
        not has_doc_access(document, collaborator)
    ):
        # Tagged user has no access to document and will therefore not be
        # notified
        return JsonResponse(
            response,
            status=200
        )
    commentator = request.user.readable_name
    collaborator_name = collaborator.readable_name
    collaborator_email = collaborator.email
    document_title = document.title
    if len(document_title) == 0:
        document_title = _('Untitled')
    link = HttpRequest.build_absolute_uri(request, document.get_absolute_url())

    if notification_type == 'mention':

        message_text = _(
            ('Hey %(collaborator_name)s,\n%(commentator)s has mentioned you '
             'in a comment in the document \'%(document)s\':'
             '\n\n%(comment_text)s'
             '\n\nGo to the document here: %(link)s')
        ) % {
               'commentator': commentator,
               'collaborator_name': collaborator_name,
               'link': link,
               'document': document_title,
               'comment_text': comment_text
        }

        body_html_title = _(
            ('Hey %(collaborator_name)s,<br>%(commentator)s has mentioned '
             'you in a comment in the document \'%(document_title)s\'.')
        ) % {
            'commentator': commentator,
            'collaborator_name': collaborator_name,
            'document_title': document_title
        }
        message_title = _('Comment on :') + ' ' + document_title
    else:
        message_text = _(
            ('Hey %(collaborator_name)s,\n%(commentator)s has assigned you to '
             'a comment in the document \'%(document)s\':\n\n%(comment_text)s'
             '\n\nGo to the document here: %(link)s')
        ) % {
               'commentator': commentator,
               'collaborator_name': collaborator_name,
               'link': link,
               'document': document_title,
               'comment_text': comment_text
        }
        body_html_title = _(
            ('Hey %(collaborator_name)s,<br>%(commentator)s has assigned you '
             'to a comment in the document \'%(document_title)s\'.')
        ) % {
            'commentator': commentator,
            'collaborator_name': collaborator_name,
            'document_title': document_title
        }
        message_title = _('Comment assignment on :') + ' ' + document_title

    body_html = _(
        ('<p>Hey %(collaborator_name)s,<br>%(commentator)s has assigned '
         'you to a comment in the document \'%(document)s\':</p>'
         '%(comment_html)s'
         '<p>Go to the document <a href="%(link)s">here</a>.</p>')
    ) % {
        'commentator': commentator,
        'collaborator_name': collaborator_name,
        'link': link,
        'document': document_title,
        'comment_html': comment_html
    }

    body_html = (
        '<h1>%(body_html_title)s</h1>'
        '<table>'
        '<tr><td>'
        '%(Document)s'
        '</td><td>'
        '<b>%(document_title)s</b>'
        '</td></tr>'
        '<tr><td>'
        '%(Author)s'
        '</td><td>'
        '%(commentator)s'
        '</td></tr>'
        '<tr><td>'
        '%(Comment)s'
        '</td><td>'
        '%(comment_html)s'
        '</td></tr>'
        '</table>'
        '<div class="actions"><a class="button" href="%(link)s">'
        '%(AccessTheDocument)s'
        '</a></div>'
    ) % {
        'body_html_title': body_html_title,
        'Document': _('Document'),
        'document_title': document_title,
        'Author': _('Author'),
        'commentator': commentator,
        'Comment': _('Comment'),
        'comment_html': comment_html,
        'link': link,
        'AccessTheDocument': _('Access the document')
    }

    send_mail(
        message_title,
        message_text,
        settings.DEFAULT_FROM_EMAIL,
        [collaborator_email],
        fail_silently=True,
        html_message=html_email(body_html)
    )
    return JsonResponse(
        response,
        status=200
    )


# maintenance views
@staff_member_required
@ajax_required
@require_POST
def get_all_old_docs(request):
    response = {}
    status = 200
    doc_list = Document.objects.filter(
        doc_version__lt=str(FW_DOCUMENT_VERSION)
    )[:10]
    response['docs'] = serializers.serialize(
        'json',
        doc_list
    )
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def save_doc(request):
    response = {}
    status = 200
    doc_id = request.POST['id']
    doc = Document.objects.get(pk=int(doc_id))
    # Only looking at fields that may have changed.
    contents = request.POST.get('contents', False)
    bibliography = request.POST.get('bibliography', False)
    comments = request.POST.get('comments', False)
    last_diffs = request.POST.get('last_diffs', False)
    version = request.POST.get('version', False)
    if contents:
        doc.contents = contents
    if bibliography:
        doc.bibliography = bibliography
    if comments:
        doc.comments = comments
    if version:
        doc.version = version
    if last_diffs:
        doc.last_diffs = last_diffs
    doc.doc_version = FW_DOCUMENT_VERSION
    doc.save()
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def get_user_biblist(request):
    response = {}
    status = 200
    user_id = request.POST['user_id']
    response['bibList'] = serializer.serialize(
        Entry.objects.filter(
            entry_owner_id=user_id
        ), fields=(
                'entry_key',
                'entry_owner',
                'bib_type',
                'fields'
        )
    )
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def get_all_template_ids(request):
    response = {}
    status = 200
    templates = DocumentTemplate.objects.filter(
        doc_version__lt=str(FW_DOCUMENT_VERSION)
    ).only('id')
    response["template_ids"] = []
    for template in templates:
        response["template_ids"].append(template.id)
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def get_template(request):
    response = {}
    status = 405
    template_id = request.POST['id']
    template = DocumentTemplate.objects.filter(pk=int(template_id)).first()
    if template:
        status = 200
        response['definition'] = template.definition
        response['title'] = template.title
        response['doc_version'] = template.doc_version
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def get_template_extras(request):
    id = request.POST['id']
    doc_template = DocumentTemplate.objects.filter(
        id=id
    ).first()
    status = 200
    if doc_template is None:
        return JsonResponse({}, status=405)
    serializer = PythonWithURLSerializer()
    export_templates = serializer.serialize(
        doc_template.exporttemplate_set.all()
    )
    document_styles = serializer.serialize(
        doc_template.documentstyle_set.all(),
        use_natural_foreign_keys=True,
        fields=['title', 'slug', 'contents', 'documentstylefile_set']
    )
    response = {
        'export_templates': export_templates,
        'document_styles': document_styles,
    }
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def save_template(request):
    response = {}
    status = 405
    template_id = request.POST['id']
    template = DocumentTemplate.objects.filter(pk=int(template_id)).first()
    if template:
        status = 200
        # Only looking at fields that may have changed.
        definition = request.POST.get('definition', False)
        if definition:
            template.definition = definition
        template.doc_version = FW_DOCUMENT_VERSION
        template.save()
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def get_all_revision_ids(request):
    response = {}
    status = 200
    revisions = DocumentRevision.objects.filter(
        doc_version__lt=str(FW_DOCUMENT_VERSION)
    ).only('id')
    response["revision_ids"] = []
    for revision in revisions:
        response["revision_ids"].append(revision.id)
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def update_revision(request):
    response = {}
    status = 405
    revision_id = request.POST['id']
    revision = DocumentRevision.objects.filter(pk=int(revision_id)).first()
    if revision:
        status = 200
        # keep the filename
        file_name = revision.file_object.name.split('/')[-1]
        # Delete the FieldFile as otherwise the file remains.
        revision.file_object.delete()
        revision.file_object = request.FILES['file']
        revision.file_object.name = file_name
        revision.doc_version = FW_DOCUMENT_VERSION
        revision.save()
    return JsonResponse(
        response,
        status=status
    )


@staff_member_required
@ajax_required
@require_POST
def add_images_to_doc(request):
    response = {}
    status = 201
    doc_id = request.POST['doc_id']
    doc = Document.objects.get(id=doc_id)
    # Delete all existing image links
    DocumentImage.objects.filter(
        document_id=doc_id
    ).delete()
    ids = request.POST.getlist('ids[]')
    for id in ids:
        doc_image_data = {
            'document': doc,
            'title': 'Deleted'
        }
        image = Image.objects.filter(id=id).first()
        if image:
            user_image = image.userimage_set.all().first()
            if user_image:
                doc_image_data['title'] = user_image.title
                doc_image_data['copyright'] = user_image.copyright
        else:
            image = Image()
            image.pk = id
            image.uploader = doc.owner
            f = open(os.path.join(
                settings.PROJECT_PATH, "base/static/img/error.png"
            ))
            image.image.save('error.png', File(f))
            image.save()
        doc_image_data['image'] = image
        DocumentImage.objects.create(**doc_image_data)
    return JsonResponse(
        response,
        status=status
    )



from .models import Document
from django.db.models.signals import post_delete
from django.dispatch import receiver


@receiver(post_delete, sender=Document)
def delete_unused_template(sender, instance, **kwargs):
    if (
        instance.template.user and
        instance.template.document_set.count() == 0 and
        instance.template.auto_delete
    ):
        # User's document template no longer used.
        instance.template.delete()



import unittest
import os
import multiprocessing
import time
import shutil
import subprocess
import urllib.request

from pathlib import Path
from tempfile import mkdtemp
import socket
import errno

from django.conf import settings

START_PORT = 8000
END_PORT = 8500


class ManageTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.port = START_PORT
        while cls.port < END_PORT:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                sock.bind(("127.0.0.1", cls.port))
            except socket.error as err:
                if err.errno == errno.EADDRINUSE:
                    cls.port += 1
                else:
                    raise err
            sock.close()
            break

    def test_startserver(self):
        os.chdir(settings.PROJECT_PATH)
        temp_dir = mkdtemp()
        p1 = multiprocessing.Process(
            target=self.start_fidus,
            args=(temp_dir,)
        )
        p1.start()
        time.sleep(1)
        sql_file = Path(os.path.join(temp_dir, 'fiduswriter.sql'))
        assert sql_file.exists()
        # Get page during transpile to see if we get setup page
        page = urllib.request.urlopen(
            'http://localhost:{}/'.format(self.port)
        ).read().decode('utf-8')
        assert "Fidus Writer is currently being updated." in page
        # We remove the project dir. The process should then finish as well.
        shutil.rmtree(temp_dir)
        p1.join()

    def start_fidus(self, temp_dir):
        os.environ.pop("PROJECT_PATH")
        if os.getenv("COVERAGE_RCFILE"):
            os.environ['COVERAGE_PROCESS_START'] = '.coveragerc'
        subprocess.call(
            [
                "./manage.py",
                "--pythonpath",
                temp_dir,
                "runserver",
                str(self.port)
            ],
            stdout=open(os.devnull, 'w'),
            stderr=subprocess.STDOUT
        )



from pathlib import Path
import shutil
from subprocess import call

from django.core.management.base import BaseCommand
from django.core.management import call_command
from django.conf import settings


BABEL_CONF = '''
module.exports = {
    presets: [
        [
            '@babel/preset-env',
            {
                targets: {
                    node: 'current',
                },
            },
        ],
    ],
}
'''


class Command(BaseCommand):
    help = 'Run jest unit tests.'

    def handle(self, *ars, **options):
        call_command('transpile')
        p = Path(settings.PROJECT_PATH) / '.transpile'
        shutil.os.chdir(p)
        conf_file = p / 'babel.config.js'

        if not conf_file.exists():
            print(f'Creating "babel.config.js" at {p}.')
            conf_file.write_text(BABEL_CONF)

        command_array = [
            p / 'node_modules' / '.bin' / 'jest',
            '--no-cache',
            '--passWithNoTests',
        ]
        return_value = call(command_array)
        if return_value > 0:
            exit(return_value)



from builtins import map
from builtins import range
import errno
import os
import socket
import sys
import threading

from django.core.exceptions import ImproperlyConfigured
from django.db import connections
from django.utils import six
from django.test.testcases import TransactionTestCase

from tornado.ioloop import IOLoop
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
from asyncio import set_event_loop_policy

from base.servers.tornado_django_hybrid import make_tornado_server


set_event_loop_policy(AnyThreadEventLoopPolicy())


class LiveTornadoThread(threading.Thread):
    """
    Thread for running a live http server while the tests are running.
    """

    def __init__(self,
                 host,
                 possible_ports,
                 static_handler,
                 connections_override=None):
        self.host = host
        self.port = None
        self.possible_ports = possible_ports
        self.is_ready = threading.Event()
        self.error = None
        self.static_handler = static_handler
        self.connections_override = connections_override
        super().__init__()

    def run(self):
        """
        Sets up the live server and databases, and then loops over handling
        http requests.
        """
        if self.connections_override:
            # Override this thread's database connections with the ones
            # provided by the main thread.
            for alias, conn in list(self.connections_override.items()):
                connections[alias] = conn
        try:
            self.httpd = make_tornado_server()

            # Go through the list of possible ports, hoping that we can find
            # one that is free to use for the WSGI server.
            for index, port in enumerate(self.possible_ports):
                try:
                    self.httpd.listen(int(port))
                except socket.error as e:
                    if (index + 1 < len(self.possible_ports) and
                            e.errno == errno.EADDRINUSE):
                        # This port is already in use, so we go on and try with
                        # the next one in the list.
                        continue
                    else:
                        # Either none of the given ports are free or the error
                        # is something else than "Address already in use". So
                        # we let that error bubble up to the main thread.
                        raise
                else:
                    # A free port was found.
                    self.port = port
                    break

            self.is_ready.set()
            self.ioloop = IOLoop.current()
            self.ioloop.start()
        except Exception as e:
            self.error = e
            self.is_ready.set()
            raise

    def terminate(self):
        if hasattr(self, 'httpd'):
            self.httpd.stop()

        if hasattr(self, 'ioloop'):
            self.ioloop.stop()


class LiveTornadoTestCase(TransactionTestCase):
    """
    Does basically the same as TransactionTestCase but also launches a live
    http server in a separate thread so that the tests may use another testing
    framework, such as Selenium for example, instead of the built-in dummy
    client.
    Note that it inherits from TransactionTestCase instead of TestCase because
    the threads do not share the same transactions (unless if using in-memory
    sqlite) and each thread needs to commit all their transactions so that the
    other thread can see the changes.
    """

    static_handler = None
    live_server_url = None

    @classmethod
    def setUpClass(cls):
        connections_override = {}
        for conn in connections.all():
            # If using in-memory sqlite databases, pass the connections to
            # the server thread.
            if (conn.vendor == 'sqlite' and
                    conn.settings_dict['NAME'] == ':memory:'):
                # Explicitly enable thread-shareability for this connection
                conn.allow_thread_sharing = True
                connections_override[conn.alias] = conn

        # Launch the live server's thread
        specified_address = os.environ.get(
            'DJANGO_LIVE_TEST_SERVER_ADDRESS', 'localhost:8081-8140')
        # The specified ports may be of the form '8000-8010,8080,9200-9300'
        # i.e. a comma-separated list of ports or ranges of ports, so we break
        # it down into a detailed list of all possible ports.
        possible_ports = []
        try:
            host, port_ranges = specified_address.split(':')
            for port_range in port_ranges.split(','):
                # A port range can be of either form: '8000' or '8000-8010'.
                extremes = list(map(int, port_range.split('-')))
                assert len(extremes) in [1, 2]
                if len(extremes) == 1:
                    # Port range of the form '8000'
                    possible_ports.append(extremes[0])
                else:
                    # Port range of the form '8000-8010'
                    for port in range(extremes[0], extremes[1] + 1):
                        possible_ports.append(port)
        except Exception:
            msg = 'Invalid address ("%s") for live server.' % specified_address
            six.reraise(
                ImproperlyConfigured,
                ImproperlyConfigured(msg),
                sys.exc_info()[2]
            )
        cls.server_thread = LiveTornadoThread(
            host,
            possible_ports,
            cls.static_handler,
            connections_override=connections_override
        )
        cls.server_thread.daemon = True
        cls.server_thread.start()

        # Wait for the live server to be ready
        cls.server_thread.is_ready.wait()
        if cls.server_thread.error:
            # Clean up behind ourselves, since tearDownClass won't get called
            # in case of errors.
            cls._tearDownClassInternal()
            raise cls.server_thread.error

        cls.live_server_url = 'http://%s:%s' % (
            cls.server_thread.host, cls.server_thread.port)

        super().setUpClass()

    @classmethod
    def _tearDownClassInternal(cls):
        # There may not be a 'server_thread' attribute if setUpClass() for some
        # reasons has raised an exception.
        if hasattr(cls, 'server_thread'):
            # Terminate the live server's thread
            cls.server_thread.terminate()
        # Restore sqlite connections' non-shareability
        for conn in connections.all():
            if (conn.vendor == 'sqlite' and
                    conn.settings_dict['NAME'] == ':memory:'):
                conn.allow_thread_sharing = False

    @classmethod
    def tearDownClass(cls):
        cls._tearDownClassInternal()
        super().tearDownClass()



# Generated by Django 2.2.9 on 2020-02-19 13:48

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import document.models
import uuid


class Migration(migrations.Migration):

    replaces = [('document', '0001_squashed_0032_document_listed'), ('document', '0002_auto_20180814_1946'), ('document', '0003_auto_20181115_0926'), ('document', '0004_documenttemplate'), ('document', '0005_document_template'), ('document', '0006_auto_20190118_1342'), ('document', '0007_auto_20190227_2105'), ('document', '0008_documenttemplate_doc_version'), ('document', '0009_documentrevision_doc_version'), ('document', '0008_auto_20190701_0943'), ('document', '0010_merge_20190701_1737'), ('document', '0011_auto_20190701_1744'), ('document', '0012_auto_20190704_1219'), ('document', '0013_auto_20190808_1126'), ('document', '0014_auto_20190811_1204'), ('document', '0015_auto_20190828_1108'), ('document', '0016_auto_20190829_2147'), ('document', '0017_remove_documenttemplate_citation_styles'), ('document', '0018_documenttemplate_import_id'), ('document', '0019_remove_documenttemplate_definition_hash')]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(blank=True, default='', max_length=255)),
                ('definition', models.TextField(default='{}')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('doc_version', models.DecimalField(decimal_places=1, default=3.1, max_digits=3)),
                ('added', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('import_id', models.CharField(blank=True, default='', max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Document',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(blank=True, default='', max_length=255)),
                ('contents', models.TextField(default='{}')),
                ('added', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='owner', to=settings.AUTH_USER_MODEL)),
                ('last_diffs', models.TextField(default='[]')),
                ('version', models.PositiveIntegerField(default=0)),
                ('comments', models.TextField(default='{}')),
                ('doc_version', models.DecimalField(decimal_places=1, default=3.1, max_digits=3)),
                ('bibliography', models.TextField(default='{}')),
                ('listed', models.BooleanField(default=True)),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='document.DocumentTemplate')),
            ],
            options={
                'ordering': ['-id'],
            },
        ),
        migrations.CreateModel(
            name='AccessRight',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('rights', models.CharField(choices=[('write', 'Writer'), ('write-tracked', 'Write with tracked changes'), ('comment', 'Commentator'), ('review', 'Reviewer'), ('read', 'Reader'), ('read-without-comments', 'Reader without comment access')], max_length=21)),
                ('document', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='document.Document')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('document', 'user')},
            },
        ),
        migrations.CreateModel(
            name='AccessRightInvite',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('email', models.EmailField(max_length=254)),
                ('rights', models.CharField(choices=[('write', 'Writer'), ('write-tracked', 'Write with tracked changes'), ('comment', 'Commentator'), ('review', 'Reviewer'), ('read', 'Reader'), ('read-without-comments', 'Reader without comment access')], max_length=21)),
                ('document', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='document.Document')),
            ],
        ),
        migrations.CreateModel(
            name='DocumentRevision',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('note', models.CharField(blank=True, default='', max_length=255)),
                ('date', models.DateTimeField(auto_now=True)),
                ('file_object', models.FileField(upload_to=document.models.revision_filename)),
                ('file_name', models.CharField(blank=True, default='', max_length=255)),
                ('document', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='document.Document')),
                ('doc_version', models.DecimalField(decimal_places=1, default=3.1, max_digits=3)),
            ],
        ),
    ]



from time import mktime
import json

from django.contrib.auth.decorators import login_required
from django.http import JsonResponse
from django.core.serializers.python import Serializer
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST

from base.decorators import ajax_required
from usermedia.models import Image, ImageCategory, UserImage
from .models import ALLOWED_FILETYPES


class SimpleSerializer(Serializer):

    def end_object(self, obj):
        self._current['id'] = obj._get_pk_val()
        self.objects.append(self._current)


serializer = SimpleSerializer()


# save changes or create a new entry
@login_required
@ajax_required
@require_POST
def save(request):
    response = {}
    response['errormsg'] = {}
    if 'image' in request.FILES and \
            request.FILES['image'].content_type not in ALLOWED_FILETYPES:
        status = 200  # Not implemented
        response['errormsg']['error'] = _('Filetype not supported')
    else:
        image = False
        if 'id' in request.POST and 'image' not in request.FILES:
            user_image = UserImage.objects.filter(
                image_id=int(request.POST['id']),
                owner=request.user
            ).first()
            if user_image:
                image = user_image.image
                status = 200
        if image is False:
            image = Image()
            image.uploader = request.user
            user_image = UserImage()
            user_image.owner = request.user
            status = 201
            if 'checksum' in request.POST:
                image.checksum = request.POST['checksum']
        user_image.title = request.POST['title']
        if 'copyright' in request.POST:
            user_image.copyright = request.POST['copyright']
        if 'cats' in request.POST:
            user_image.image_cat = request.POST['cats']
        if 'image' in request.FILES:
            image.image = request.FILES['image']
        if status == 201 and 'image' not in request.FILES:
            status = 200
            response['errormsg']['error'] = _('No file uploaded')
        else:
            image.save()
            user_image.image = image
            user_image.save()
            response['values'] = {
                'id': image.id,
                'title': user_image.title,
                'copyright': json.loads(user_image.copyright),
                'image': image.image.url,
                'file_type': image.file_type,
                'added': mktime(image.added.timetuple()) * 1000,
                'checksum': image.checksum,
                'cats': json.loads(user_image.image_cat)
            }
            if image.thumbnail:
                response['values']['thumbnail'] = image.thumbnail.url
                response['values']['height'] = image.height
                response['values']['width'] = image.width
    return JsonResponse(
        response,
        status=status
    )


# delete an image
@login_required
@ajax_required
@require_POST
def delete(request):
    response = {}
    status = 201
    ids = request.POST.getlist('ids[]')
    UserImage.objects.filter(
        image_id__in=ids,
        owner=request.user
    ).delete()
    for image in Image.objects.filter(id__in=ids):
        if image.is_deletable():
            image.delete()
    return JsonResponse(
        response,
        status=status
    )


# returns list of images
@login_required
@ajax_required
@require_POST
def images(request):
    response = {}
    status = 200
    response['imageCategories'] = serializer.serialize(
        ImageCategory.objects.filter(category_owner=request.user))
    response['images'] = []
    user_images = UserImage.objects.filter(owner=request.user)
    for user_image in user_images:
        image = user_image.image
        if image.image:
            field_obj = {
                'id': image.id,
                'title': user_image.title,
                'copyright': json.loads(user_image.copyright),
                'image': image.image.url,
                'file_type': image.file_type,
                'added': mktime(image.added.timetuple()) * 1000,
                'checksum': image.checksum,
                'cats': json.loads(user_image.image_cat)
            }
            if image.thumbnail:
                field_obj['thumbnail'] = image.thumbnail.url
                field_obj['height'] = image.height
                field_obj['width'] = image.width
            response['images'].append(field_obj)
    return JsonResponse(
        response,
        status=status
    )


# save changes or create a new category
@login_required
@ajax_required
@require_POST
def save_category(request):
    response = {}
    response['entries'] = []
    ids = request.POST.getlist('ids[]')
    titles = request.POST.getlist('titles[]')
    ImageCategory.objects.filter(
        category_owner=request.user
    ).exclude(id__in=ids).delete()
    x = 0
    for the_id in ids:
        the_id = int(the_id)
        the_title = titles[x]
        x += 1
        if 0 == the_id:
            # if the category is new, then create new
            the_cat = ImageCategory(
                category_title=the_title,
                category_owner=request.user)
        else:
            # if the category already exists, update the title
            the_cat = ImageCategory.objects.get(pk=the_id)
            the_cat.category_title = the_title
        the_cat.save()
        response['entries'].append(
            {'id': the_cat.id, 'category_title': the_cat.category_title})
    status = 201

    return JsonResponse(
        response,
        status=status
    )



"""
Application that contains main app settings
(Hosts, middleware, templates, installed apps)
Urls to route between applications,
statics that are shared among different django applications
Also contains helper for starting tornado server and management commands
"""
default_app_config = 'base.apps.BaseConfig'



from django.conf.urls import url

from . import views

urlpatterns = [
    url(
        '^delete_document_style/$',
        views.delete_document_style,
        name='delete_document_style'
    ),
    url(
        '^save_document_style/$',
        views.save_document_style,
        name='save_document_style'
    ),
    url(
        '^delete_export_template/$',
        views.delete_export_template,
        name='delete_export_template'
    ),
    url(
        '^save_export_template/$',
        views.save_export_template,
        name='save_export_template'
    ),
]



from django.shortcuts import render
from django.views.decorators.csrf import ensure_csrf_cookie
from django.contrib.sites.shortcuts import get_current_site
from django.contrib.flatpages.models import FlatPage
from django.http import JsonResponse
from django.views.decorators.http import require_POST

from allauth.socialaccount.models import providers

from user import util as userutil

from .decorators import ajax_required


@ensure_csrf_cookie
def app(request):
    """
    Load a page controlled by the JavaScript app.
    Used all user facing pages after login.
    """
    return render(request, 'app.html')


def api_404(request):
    """
    Show a 404 error within the API.
    """
    return render(request, 'api_404.html', status=404)


@ajax_required
@require_POST
def configuration(request):
    """
    Load the configuration options of the page that are request dependent.
    """
    socialaccount_providers = []
    for provider in providers.registry.get_list():
        socialaccount_providers.append({
            'id': provider.id,
            'name': provider.name,
            'login_url': provider.get_login_url(request)
        })
    response = {
        'language': request.LANGUAGE_CODE,
        'socialaccount_providers': socialaccount_providers
    }
    if request.user.is_authenticated:
        response['user'] = {
            'id': request.user.id,
            'username': request.user.username,
            'first_name': request.user.first_name,
            'name': request.user.readable_name,
            'last_name': request.user.last_name,
            'avatar': userutil.get_user_avatar_url(request.user),
            'emails': [],
            'socialaccounts': [],
            'is_authenticated': True
        }

        for emailaddress in request.user.emailaddress_set.all():
            email = {
                'address': emailaddress.email,
            }
            if emailaddress.primary:
                email['primary'] = True
            if emailaddress.verified:
                email['verified'] = True
            response['user']['emails'].append(email)
        for account in request.user.socialaccount_set.all():
            provider_account = account.get_provider_account()
            response['user']['socialaccounts'].append({
                'id': account.id,
                'provider': account.provider,
                'name': provider_account.to_str()
            })

    else:
        response['user'] = {
            'is_authenticated': False
        }
    return JsonResponse(
        response,
        status=200
    )


def manifest_json(request):
    """
    Load the manifest.json.
    """
    return render(request, 'manifest.json')


# view is shown only in admin interface, so authentication is taken care of
def admin_console(request):
    """
    Load the admin console page.
    """
    return render(request, 'admin/console.html')


@ajax_required
@require_POST
def flatpage(request):
    """
    Models: `flatpages.flatpages`
    Context:
        flatpage
            `flatpages.flatpages` object
    """
    response = {}
    status = 404
    url = request.POST['url']
    site_id = get_current_site(request).id
    flatpage = FlatPage.objects.filter(url=url, sites=site_id).first()
    if flatpage:
        status = 200
        response['title'] = flatpage.title
        response['content'] = flatpage.content
    return JsonResponse(
        response,
        status=status
    )



# code adapted by github.com/jobdiogenes from
# https://github.com/dkarchmer/aws-eb-docker-django/blob/master
# /authentication/manage
# used to help automation install like in docker.
# Create admins accounts if no users exists.
# Password 'admin' is used unless defined by ADMIN_PASSWORD
from django.conf import settings
from django.core.management.base import BaseCommand
from django.contrib.auth.models import User
from os import getenv


class Command(BaseCommand):

    def handle(self, *args, **options):
        if User.objects.count() == 0:
            for user in settings.ADMINS:
                username = user[0].replace(' ', '')
                email = user[1]
                if getenv('ADMIN_PASSWORD') != '':
                    password = getenv('ADMIN_PASSWORD')
                else:
                    password = 'admin'
                print('Creating account for %s (%s)' % (username, email))
                admin = User.objects.create_superuser(
                    username=username,
                    email=email,
                    password=password
                )
                admin.is_active = True
                admin.is_admin = True
                admin.save()
        else:
            print(
                'Admin accounts can only be initialized if no Accounts exist'
            )



from builtins import str
import uuid
import atexit
import json
from time import mktime, time
from copy import deepcopy

from jsonpatch import apply_patch, JsonPatchConflict, JsonPointerException

from django.db.utils import DatabaseError
from document.helpers.session_user_info import SessionUserInfo
from document.helpers.serializers import PythonWithURLSerializer
from base.ws_handler import BaseWebSocketHandler
import logging
from tornado.escape import json_decode, json_encode
from document.models import COMMENT_ONLY, CAN_UPDATE_DOCUMENT, \
    CAN_COMMUNICATE, FW_DOCUMENT_VERSION, DocumentTemplate, Document
from usermedia.models import Image, DocumentImage, UserImage
from user.util import get_user_avatar_url

from django.db.models import F, Q

logger = logging.getLogger(__name__)


class WebSocket(BaseWebSocketHandler):
    sessions = dict()
    history_length = 1000  # Only keep the last 1000 diffs

    def open(self, arg):
        super().open(arg)
        if len(self.args) < 2:
            self.access_denied()
            return
        self.document_id = int(self.args[0])

    def confirm_diff(self, rid):
        response = {
            'type': 'confirm_diff',
            'rid': rid
        }
        self.send_message(response)

    def subscribe_doc(self, connection_count=0):
        self.user_info = SessionUserInfo(self.user)
        doc_db, can_access = self.user_info.init_access(
            self.document_id
        )
        if not can_access or float(doc_db.doc_version) != FW_DOCUMENT_VERSION:
            self.access_denied()
            return
        if (
            doc_db.id in WebSocket.sessions and
            len(WebSocket.sessions[doc_db.id]['participants']) > 0
        ):
            logger.debug("Serving already opened file")
            self.doc = WebSocket.sessions[doc_db.id]
            self.id = max(self.doc['participants']) + 1
            self.doc['participants'][self.id] = self
            logger.debug("id when opened %s" % self.id)
        else:
            logger.debug("Opening file")
            self.id = 0
            self.doc = {
                'db': doc_db,
                'participants': {
                    0: self
                },
                'last_diffs': json_decode(doc_db.last_diffs),
                'comments': json_decode(doc_db.comments),
                'bibliography': json_decode(doc_db.bibliography),
                'contents': json_decode(doc_db.contents),
                'version': doc_db.version,
                'title': doc_db.title,
                'id': doc_db.id,
                'template': {
                    'id': doc_db.template.id,
                    'definition': json_decode(doc_db.template.definition)
                }
            }
            WebSocket.sessions[doc_db.id] = self.doc
        self.send_message({
            'type': 'subscribed'
        })
        if connection_count < 1:
            self.send_styles()
            self.send_document()
        if self.can_communicate():
            self.handle_participant_update()

    def send_styles(self):
        doc_db = self.doc['db']
        response = dict()
        response['type'] = 'styles'
        serializer = PythonWithURLSerializer()
        export_temps = serializer.serialize(
            doc_db.template.exporttemplate_set.all(),
            fields=['file_type', 'template_file', 'title']
        )
        document_styles = serializer.serialize(
            doc_db.template.documentstyle_set.all(),
            use_natural_foreign_keys=True,
            fields=['title', 'slug', 'contents', 'documentstylefile_set']
        )
        document_templates = {}
        for obj in DocumentTemplate.objects.filter(
            Q(user=self.user) | Q(user=None)
        ).order_by(F('user').desc(nulls_first=True)):
            document_templates[obj.import_id] = {
                'title': obj.title,
                'id': obj.id
            }

        response['styles'] = {
            'export_templates': [obj['fields'] for obj in export_temps],
            'document_styles': [obj['fields'] for obj in document_styles],
            'document_templates': document_templates
        }
        self.send_message(response)

    def unfixable(self):
        self.send_document()

    def send_document(self):
        response = dict()
        response['type'] = 'doc_data'
        doc_owner = self.doc['db'].owner
        response['doc_info'] = {
            'id': self.doc['id'],
            'is_owner': self.user_info.is_owner,
            'access_rights': self.user_info.access_rights,
            'owner': {
                'id': doc_owner.id,
                'name': doc_owner.readable_name,
                'username': doc_owner.username,
                'avatar': get_user_avatar_url(doc_owner),
                'team_members': []
            }
        }
        response['doc'] = {
            'v': self.doc['version'],
            'contents': self.doc['contents'],
            'bibliography': self.doc['bibliography'],
            'template': self.doc['template'],
            'images': {}
        }
        response['time'] = int(time()) * 1000
        for dimage in DocumentImage.objects.filter(document_id=self.doc['id']):
            image = dimage.image
            field_obj = {
                'id': image.id,
                'title': dimage.title,
                'copyright': json.loads(dimage.copyright),
                'image': image.image.url,
                'file_type': image.file_type,
                'added': mktime(image.added.timetuple()) * 1000,
                'checksum': image.checksum,
                'cats': []
            }
            if image.thumbnail:
                field_obj['thumbnail'] = image.thumbnail.url
                field_obj['height'] = image.height
                field_obj['width'] = image.width
            response['doc']['images'][image.id] = field_obj
        if self.user_info.access_rights == 'read-without-comments':
            response['doc']['comments'] = []
        elif self.user_info.access_rights == 'review':
            # Reviewer should only get his/her own comments
            filtered_comments = {}
            for key, value in list(self.doc["comments"].items()):
                if value["user"] == self.user_info.user.id:
                    filtered_comments[key] = value
            response['doc']['comments'] = filtered_comments
        else:
            response['doc']['comments'] = self.doc["comments"]
        for team_member in doc_owner.leader.all():
            tm_object = dict()
            tm_object['id'] = team_member.member.id
            tm_object['name'] = team_member.member.readable_name
            tm_object['username'] = team_member.member.get_username()
            tm_object['avatar'] = get_user_avatar_url(team_member.member)
            response['doc_info']['owner']['team_members'].append(tm_object)
        response['doc_info']['session_id'] = self.id
        self.send_message(response)

    def reject_message(self, message):
        if (message["type"] == "diff"):
            self.send_message({
                'type': 'reject_diff',
                'rid': message['rid']
            })

    def handle_message(self, message):
        if message["type"] == 'subscribe':
            connection_count = 0
            if 'connection' in message:
                connection_count = message['connection']
            self.subscribe_doc(connection_count)
            return
        if self.user_info.document_id not in WebSocket.sessions:
            logger.debug('receiving message for closed document')
            return
        if message["type"] == 'get_document':
            self.send_document()
        elif (
            message["type"] == 'participant_update' and
            self.can_communicate()
        ):
            self.handle_participant_update()
        elif message["type"] == 'chat' and self.can_communicate():
            self.handle_chat(message)
        elif message["type"] == 'check_version':
            self.check_version(message)
        elif message["type"] == 'selection_change':
            self.handle_selection_change(message)
        elif message["type"] == 'diff' and self.can_update_document():
            self.handle_diff(message)

    def update_bibliography(self, bibliography_updates):
        for bu in bibliography_updates:
            if "id" not in bu:
                continue
            id = bu["id"]
            if bu["type"] == "update":
                self.doc["bibliography"][id] = bu["reference"]
            elif bu["type"] == "delete":
                del self.doc["bibliography"][id]

    def update_images(self, image_updates):
        for iu in image_updates:
            if "id" not in iu:
                continue
            id = iu["id"]
            if iu["type"] == "update":
                # Ensure that access rights exist
                if not UserImage.objects.filter(
                    image__id=id,
                    owner=self.user_info.user
                ).exists():
                    continue
                doc_image = DocumentImage.objects.filter(
                    document_id=self.doc["id"],
                    image_id=id
                ).first()
                if doc_image:
                    doc_image.title = iu["image"]["title"]
                    doc_image.copyright = json.dumps(iu["image"]["copyright"])
                    doc_image.save()
                else:
                    DocumentImage.objects.create(
                        document_id=self.doc["id"],
                        image_id=id,
                        title=iu["image"]["title"],
                        copyright=json.dumps(iu["image"]["copyright"])
                    )
            elif iu["type"] == "delete":
                DocumentImage.objects.filter(
                    document_id=self.doc["id"],
                    image_id=id
                ).delete()
                for image in Image.objects.filter(id=id):
                    if image.is_deletable():
                        image.delete()

    def update_comments(self, comments_updates):
        comments_updates = deepcopy(comments_updates)
        for cd in comments_updates:
            if "id" not in cd:
                # ignore
                continue
            id = cd["id"]
            if cd["type"] == "create":
                self.doc["comments"][id] = {
                    "user": cd["user"],
                    "username": cd["username"],
                    "assignedUser": cd["assignedUser"],
                    "assignedUsername": cd["assignedUsername"],
                    "date": cd["date"],
                    "comment": cd["comment"],
                    "isMajor": cd["isMajor"],
                    "resolved": cd["resolved"],
                }
            elif cd["type"] == "delete":
                del self.doc["comments"][id]
            elif cd["type"] == "update":
                self.doc["comments"][id]["comment"] = cd["comment"]
                if "isMajor" in cd:
                    self.doc["comments"][id][
                        "isMajor"] = cd["isMajor"]
                if "assignedUser" in cd and "assignedUsername" in cd:
                    self.doc["comments"][id][
                        "assignedUser"] = cd["assignedUser"]
                    self.doc["comments"][id][
                        "assignedUsername"] = cd["assignedUsername"]
                if "resolved" in cd:
                    self.doc["comments"][id][
                        "resolved"] = cd["resolved"]
            elif cd["type"] == "add_answer":
                if "answers" not in self.doc["comments"][id]:
                    self.doc["comments"][id]["answers"] = []
                self.doc["comments"][id]["answers"].append({
                    "id": cd["answerId"],
                    "user": cd["user"],
                    "username": cd["username"],
                    "date": cd["date"],
                    "answer": cd["answer"]
                })
            elif cd["type"] == "delete_answer":
                answer_id = cd["answerId"]
                for answer in self.doc["comments"][id]["answers"]:
                    if answer["id"] == answer_id:
                        self.doc["comments"][id]["answers"].remove(answer)
            elif cd["type"] == "update_answer":
                answer_id = cd["answerId"]
                for answer in self.doc["comments"][id]["answers"]:
                    if answer["id"] == answer_id:
                        answer["answer"] = cd["answer"]

    def handle_participant_update(self):
        WebSocket.send_participant_list(self.user_info.document_id)

    def handle_chat(self, message):
        chat = {
            "id": str(uuid.uuid4()),
            "body": message['body'],
            "from": self.user_info.user.id,
            "type": 'chat'
        }
        WebSocket.send_updates(chat, self.user_info.document_id)

    def handle_selection_change(self, message):
        if self.user_info.document_id in WebSocket.sessions and message[
                "v"] == self.doc['version']:
            WebSocket.send_updates(
                message, self.user_info.document_id, self.id)

    # Checks if the diff only contains changes to comments.
    def only_comments(self, message):
        allowed_operations = ['addMark', 'removeMark']
        only_comment = True
        if "ds" in message:  # ds = document steps
            for step in message["ds"]:
                if not (step['stepType'] in allowed_operations and step[
                        'mark']['type'] == 'comment'):
                    only_comment = False
        return only_comment

    def handle_diff(self, message):
        pv = message["v"]
        dv = self.doc['version']
        logger.debug("PV: %d, DV: %d" % (pv, dv))
        if (
            self.user_info.access_rights in COMMENT_ONLY and
            not self.only_comments(message)
        ):
            logger.error(
                (
                    'received non-comment diff from comment-only '
                    'collaborator. Discarding.'
                )
            )
            return
        if pv == dv:
            self.doc["last_diffs"].append(message)
            self.doc["last_diffs"] = self.doc[
                "last_diffs"
            ][-self.history_length:]
            self.doc['version'] += 1
            if "jd" in message:  # jd = json diff
                try:
                    apply_patch(
                       self.doc['contents'],
                       message["jd"],
                       True
                    )
                except (JsonPatchConflict, JsonPointerException):
                    logger.exception("Cannot apply json diff.")
                    logger.error(json_encode(message))
                    logger.error(json_encode(self.doc['contents']))
                    self.unfixable()
                # The json diff is only needed by the python backend which does
                # not understand the steps. It can therefore be removed before
                # broadcast to other clients.
                del message["jd"]
            if "ti" in message:  # ti = title
                self.doc["title"] = message["ti"]
            if "cu" in message:  # cu = comment updates
                self.update_comments(message["cu"])
            if "bu" in message:  # bu = bibliography updates
                self.update_bibliography(message["bu"])
            if "iu" in message:  # iu = image updates
                self.update_images(message["iu"])
            if self.doc['version'] % 10 == 0:
                WebSocket.save_document(self.user_info.document_id)
            self.confirm_diff(message["rid"])
            WebSocket.send_updates(
                message,
                self.user_info.document_id,
                self.id,
                self.user_info.user.id
            )
        elif pv < dv:
            if pv + len(self.doc["last_diffs"]) >= dv:
                # We have enough last_diffs stored to fix it.
                logger.debug("can fix it")
                number_diffs = pv - dv
                messages = self.doc["last_diffs"][number_diffs:]
                for message in messages:
                    new_message = message.copy()
                    new_message["server_fix"] = True
                    self.send_message(new_message)
            else:
                logger.debug('unfixable')
                # Client has a version that is too old to be fixed
                self.unfixable()
        else:
            # Client has a higher version than server. Something is fishy!
            logger.debug('unfixable')

    def check_version(self, message):
        pv = message["v"]
        dv = self.doc['version']
        logger.debug("PV: %d, DV: %d" % (pv, dv))
        if pv == dv:
            response = {
                "type": "confirm_version",
                "v": pv,
            }
            self.send_message(response)
            return
        elif pv + len(self.doc["last_diffs"]) >= dv:
            logger.debug("can fix it")
            number_diffs = pv - dv
            messages = self.doc["last_diffs"][number_diffs:]
            for message in messages:
                new_message = message.copy()
                new_message["server_fix"] = True
                self.send_message(new_message)
            return
        else:
            logger.debug('unfixable')
            # Client has a version that is too old
            self.unfixable()
            return

    def can_update_document(self):
        return self.user_info.access_rights in CAN_UPDATE_DOCUMENT

    def can_communicate(self):
        return self.user_info.access_rights in CAN_COMMUNICATE

    def on_close(self):
        logger.debug('Websocket closing')
        if (
            hasattr(self, 'user_info') and
            hasattr(self.user_info, 'document_id') and
            self.user_info.document_id in WebSocket.sessions and
            hasattr(self, 'id') and
            self.id in WebSocket.sessions[
                self.user_info.document_id
            ]['participants']
        ):
            del self.doc['participants'][self.id]
            if len(self.doc['participants']) == 0:
                WebSocket.save_document(self.user_info.document_id)
                del WebSocket.sessions[self.user_info.document_id]
                logger.debug("noone left")
            else:
                WebSocket.send_participant_list(self.user_info.document_id)

    @classmethod
    def send_participant_list(cls, document_id):
        if document_id in WebSocket.sessions:
            participant_list = []
            for session_id, waiter in list(cls.sessions[
                document_id
            ]['participants'].items()):
                access_rights = waiter.user_info.access_rights
                if access_rights not in CAN_COMMUNICATE:
                    continue
                participant_list.append({
                    'session_id': session_id,
                    'id': waiter.user_info.user.id,
                    'name': waiter.user_info.user.readable_name,
                    'avatar': get_user_avatar_url(waiter.user_info.user)
                })
            message = {
                "participant_list": participant_list,
                "type": 'connections'
            }
            WebSocket.send_updates(message, document_id)

    @classmethod
    def send_updates(cls, message, document_id, sender_id=None, user_id=None):
        logger.debug(
            "Sending message to %d waiters",
            len(cls.sessions[document_id]['participants'])
        )
        for waiter in list(cls.sessions[document_id]['participants'].values()):
            if waiter.id != sender_id:
                access_rights = waiter.user_info.access_rights
                if "comments" in message and len(message["comments"]) > 0:
                    # Filter comments if needed
                    if access_rights == 'read-without-comments':
                        # The reader should not receive the comments update, so
                        # we remove the comments from the copy of the message
                        # sent to the reviewer. We still need to send the rest
                        # of the message as it may contain other diff
                        # information.
                        message = deepcopy(message)
                        message['comments'] = []
                    elif (
                        access_rights == 'review' and
                        user_id != waiter.user_info.user.id
                    ):
                        # The reviewer should not receive comments updates from
                        # others than themselves, so we remove the comments
                        # from the copy of the message sent to the reviewer
                        # that are not from them. We still need to send the
                        # rest of the message as it may contain other diff
                        # information.
                        message = deepcopy(message)
                        message['comments'] = []
                elif (
                    message['type'] in ["chat", "connections"] and
                    access_rights not in CAN_COMMUNICATE
                ):
                    continue
                elif (
                    message['type'] == "selection_change" and
                    access_rights not in CAN_COMMUNICATE and
                    user_id != waiter.user_info.user.id
                ):
                    continue
                waiter.send_message(message)

    @classmethod
    def save_document(cls, document_id):
        doc = cls.sessions[document_id]
        doc_db = doc['db']
        if doc_db.version == doc['version']:
            return
        doc_db.title = doc['title'][-255:]
        doc_db.version = doc['version']
        doc_db.contents = json_encode(doc['contents'])
        doc_db.last_diffs = json_encode(doc['last_diffs'])
        doc_db.comments = json_encode(doc['comments'])
        doc_db.bibliography = json_encode(doc['bibliography'])
        logger.debug('saving document # %d' % doc_db.id)
        logger.debug('version %d' % doc_db.version)
        try:
            # this try block is to avoid a db exception
            # in case the doc has been deleted from the db
            # in fiduswriter the owner of a doc could delete a doc
            # while an invited writer is editing the same doc
            doc_db.save(update_fields=[
                        'title',
                        'version',
                        'contents',
                        'last_diffs',
                        'comments',
                        'bibliography',
                        'updated'])
        except DatabaseError as e:
            expected_msg = 'Save with update_fields did not affect any rows.'
            if str(e) == expected_msg:
                cls.__insert_document(doc=doc_db)
            else:
                raise e

    @classmethod
    def __insert_document(cls, doc: Document) -> None:
        """
        Purpose:
        during plugin tests we experienced Integrity errors
         at the end of tests.
        This exception occurs while handling another exception,
         so in order to have a clean tests output
          we raise the exception in a way we don't output
           misleading error messages related to different exceptions

        :param doc: socket document model instance
        :return: None
        """
        from django.db.utils import IntegrityError
        try:
            doc.save()
        except IntegrityError as e:
            raise IntegrityError(
                'plugin test error when we try to save a doc already deleted '
                'along with the rest of db data so it '
                'raises an Integrity error: {}'.format(e)) from None

    @classmethod
    def save_all_docs(cls):
        for document_id in cls.sessions:
            cls.save_document(document_id)


atexit.register(WebSocket.save_all_docs)



import shutil
import os
import json
from subprocess import call

from django.apps import apps
from django.core.management.base import BaseCommand
from django.core.management import call_command
from django.conf import settings


class Command(BaseCommand):
    help = 'Check JavaScript files with ESLint'

    def add_arguments(self, parser):
        parser.add_argument(
            '--fix',
            action='store_true',
            dest='fix',
            default=False,
            help='Whether to also attempt to fix files'
        )

    def handle(self, *args, **options):
        call_command("npm_install")
        shutil.os.chdir(settings.PROJECT_PATH)
        apps_paths = []
        for app in list(apps.get_app_configs()):
            apps_paths.append(app.path)
        os.environ['DJANGO_APPS_PATHS'] = json.dumps(apps_paths)
        command_array = [
            os.path.join(
                settings.PROJECT_PATH,
                ".transpile/node_modules/.bin/eslint"
            ),
            "--max-warnings=0",
            "--ignore-path",
            os.path.join(
                settings.SRC_PATH,
                ".eslintignore"
            ),
            "-c",
            os.path.join(
                settings.SRC_PATH,
                ".eslintrc.js"
            ),
            "."
        ]
        if options['fix']:
            command_array.append('--fix')
        return_value = call(command_array)
        if return_value > 0:
            exit(return_value)



import uuid

from builtins import str
from builtins import object

from django.db import models
from django.db.utils import OperationalError, ProgrammingError
from django.contrib.auth.models import User
from django.core import checks

# FW_DOCUMENT_VERSION:
# Also defined in frontend
# document/static/js/modules/schema/index.js

FW_DOCUMENT_VERSION = 3.2


class DocumentTemplate(models.Model):
    title = models.CharField(max_length=255, default='', blank=True)
    import_id = models.CharField(max_length=255, default='', blank=True)
    definition = models.TextField(default='{}')
    doc_version = models.DecimalField(
        max_digits=3,
        decimal_places=1,
        default=FW_DOCUMENT_VERSION
    )
    user = models.ForeignKey(
        User,
        null=True,
        blank=True,
        on_delete=models.deletion.CASCADE
    )
    added = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    auto_delete = True

    def __str__(self):
        return self.title

    def is_deletable(self):
        reverse_relations = [
            f for f in self._meta.model._meta.get_fields()
            if (f.one_to_many or f.one_to_one) and
            f.auto_created and not f.concrete and
            f.name not in ['documentstyle', 'exporttemplate']
        ]

        for r in reverse_relations:
            if r.remote_field.model.objects.filter(
                **{r.field.name: self}
            ).exists():
                return False
        return True

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(cls._check_doc_versions(**kwargs))
        return errors

    @classmethod
    def _check_doc_versions(cls, **kwargs):
        try:
            if len(
                cls.objects.filter(doc_version__lt=str(FW_DOCUMENT_VERSION))
            ):
                return [
                    checks.Warning(
                        'Document templates need to be upgraded. Please '
                        'navigate to /admin/document/document/maintenance/ '
                        'with a browser as a superuser and upgrade all '
                        'document templates on this server.',
                        obj=cls
                    )
                ]
            else:
                return []
        except (ProgrammingError, OperationalError):
            # Database has not yet been initialized, so don't throw any error.
            return []


class Document(models.Model):
    title = models.CharField(max_length=255, default='', blank=True)
    contents = models.TextField(default='{}')  # json object of content
    doc_version = models.DecimalField(
        max_digits=3,
        decimal_places=1,
        default=FW_DOCUMENT_VERSION
    )
    # The doc_version is the version of the data format in the contents field.
    # We upgrade the contents field in JavaScript and not migrations so that
    # the same code can be used for migrations and for importing old fidus
    # files that are being uploaded. This field is only used for upgrading data
    # and is therefore not handed to the editor or document overview page.
    version = models.PositiveIntegerField(default=0)
    last_diffs = models.TextField(default='[]')
    # The last few diffs that were received and approved. The number of stored
    # diffs should always be equivalent to or more than all the diffs since the
    # last full save of the document.
    owner = models.ForeignKey(
        User,
        related_name='owner',
        on_delete=models.deletion.CASCADE
    )
    added = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    comments = models.TextField(default='{}')
    bibliography = models.TextField(default='{}')
    # Whether or not document is listed on document overview list page.
    # True by default and for all normal documents. Can be set to False when
    # documents are added in plugins that list these documents somewhere else.
    listed = models.BooleanField(default=True)
    template = models.ForeignKey(
        DocumentTemplate,
        on_delete=models.deletion.CASCADE
    )

    def __str__(self):
        if len(self.title) > 0:
            return self.title + ' (' + str(self.id) + ')'
        else:
            return str(self.id)

    class Meta(object):
        ordering = ['-id']

    def get_absolute_url(self):
        return "/document/%i/" % self.id

    def is_deletable(self):
        reverse_relations = [
            f for f in self._meta.model._meta.get_fields()
            if (f.one_to_many or f.one_to_one) and
            f.auto_created and not f.concrete and
            f.name not in [
                'accessright',
                'accessrightinvite',
                'documentrevision',
                'documentimage'
            ]
        ]

        for r in reverse_relations:
            if r.remote_field.model.objects.filter(
                **{r.field.name: self}
            ).exists():
                return False
        return True

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(cls._check_doc_versions(**kwargs))
        return errors

    @classmethod
    def _check_doc_versions(cls, **kwargs):
        try:
            if len(
                cls.objects.filter(doc_version__lt=str(FW_DOCUMENT_VERSION))
            ):
                return [
                    checks.Warning(
                        'Documents need to be upgraded. Please navigate to '
                        '/admin/document/document/maintenance/ with a browser '
                        'as a superuser and upgrade all documents on this '
                        'server.',
                        obj=cls
                    )
                ]
            else:
                return []
        except (ProgrammingError, OperationalError):
            # Database has not yet been initialized, so don't throw any error.
            return []


RIGHTS_CHOICES = (
    ('write', 'Writer'),
    # Can write contents and can read+write comments.
    # Can chat with collaborators.
    # Has read access to revisions.
    ('write-tracked', 'Write with tracked changes'),
    # Can write tracked contents and can read/write comments.
    # Cannot turn off tracked changes.
    # Can chat with collaborators.
    # Has read access to revisions.
    ('comment', 'Commentator'),
    # Can read contents and can read+write comments.
    # Can chat with collaborators.
    # Has read access to revisions.
    ('review', 'Reviewer'),
    # Can read the contents and can read/write his own comments.
    # Comments by users with this access right only show the user's
    # numeric ID, not their username.
    # Cannot chat with collaborators nor see that they are connected.
    # Has no access to revisions.
    ('read', 'Reader'),
    # Can read contents, including comments
    # Can chat with collaborators.
    # Has read access to revisions.
    ('read-without-comments', 'Reader without comment access'),
    # Can read contents, but not the comments.
    # Cannot chat with collaborators.
    # Has no access to revisions.
)

# Editor and Reviewer can only comment and not edit document
COMMENT_ONLY = ('review', 'comment')

CAN_UPDATE_DOCUMENT = ['write', 'write-tracked', 'review', 'comment']

# Whether the collaborator is allowed to know about other collaborators
# and communicate with them.
CAN_COMMUNICATE = ['read', 'write', 'comment', 'write-tracked']


class AccessRight(models.Model):
    document = models.ForeignKey(Document, on_delete=models.deletion.CASCADE)
    user = models.ForeignKey(User, on_delete=models.deletion.CASCADE)
    rights = models.CharField(
        max_length=21,
        choices=RIGHTS_CHOICES,
        blank=False)

    class Meta(object):
        unique_together = (("document", "user"),)

    def __str__(self):
        return (
            '%(name)s %(rights)s on %(doc_id)d' %
            {
                'name': self.user.readable_name,
                'rights': self.rights,
                'doc_id': self.document.id
            }
        )


class AccessRightInvite(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    email = models.EmailField()  # The email where the invite was sent
    document = models.ForeignKey(Document, on_delete=models.deletion.CASCADE)
    rights = models.CharField(
        max_length=21,
        choices=RIGHTS_CHOICES,
        blank=False)

    def __str__(self):
        return (
            '%(email)s %(rights)s on %(doc_id)d: %(id)d' %
            {
                'email': self.email,
                'rights': self.rights,
                'doc_id': self.document.id,
                'id': self.id
            }
        )

    def get_absolute_url(self):
        return "/invite/%i/" % self.id


def revision_filename(instance, filename):
    return "document-revisions/{id}.fidus".format(id=instance.pk)


class DocumentRevision(models.Model):
    document = models.ForeignKey(Document, on_delete=models.deletion.CASCADE)
    doc_version = models.DecimalField(
        max_digits=3,
        decimal_places=1,
        default=FW_DOCUMENT_VERSION
    )
    note = models.CharField(max_length=255, default='', blank=True)
    date = models.DateTimeField(auto_now=True)
    file_object = models.FileField(upload_to=revision_filename)
    file_name = models.CharField(max_length=255, default='', blank=True)

    def save(self, *args, **kwargs):
        if self.pk is None:
            # We remove the file_object the first time so that we can use the
            # pk as the name of the saved revision file.
            file_object = self.file_object
            self.file_object = None
            super().save(*args, **kwargs)
            self.file_object = file_object
            kwargs.pop('force_insert', None)
        super(DocumentRevision, self).save(*args, **kwargs)

    def __str__(self):
        if len(self.note) > 0:
            return self.note + ' (' + str(self.id) + ') of ' + \
                str(self.document.id)
        else:
            return str(self.id) + ' of ' + str(self.document.id)

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(cls._check_doc_versions(**kwargs))
        return errors

    @classmethod
    def _check_doc_versions(cls, **kwargs):
        try:
            if len(
                cls.objects.filter(doc_version__lt=str(FW_DOCUMENT_VERSION))
            ):
                return [
                    checks.Warning(
                        'Document revisions need to be upgraded. Please '
                        'navigate to /admin/document/document/maintenance/ '
                        'with a browser as a superuser and upgrade all '
                        'document revisions on this server.',
                        obj=cls
                    )
                ]
            else:
                return []
        except (ProgrammingError, OperationalError):
            # Database has not yet been initialized, so don't throw any error.
            return []
